# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import logging
from pathlib import Path
from statistics import mean
from typing import Dict, List, Optional

import torch
from numpy import int32, ndarray
from PIL import Image, ImageDraw, ImageFont
from PIL.ImageFont import FreeTypeFont
from prettytable import MARKDOWN, PrettyTable
from torch.optim import Adam

from dan.ocr.decoder import GlobalHTADecoder
from dan.ocr.encoder import FCN_Encoder
from dan.ocr.transforms import Preprocessing

logger = logging.getLogger(__name__)

METRICS_TABLE_HEADER = {
    "cer": "CER (HTR-NER)",
    "cer_no_token": "CER (HTR)",
    "wer": "WER (HTR-NER)",
    "wer_no_token": "WER (HTR)",
    "wer_no_punct": "WER (HTR no punct)",
    "ner": "NER",
}
REVERSE_HEADER = {column: metric for metric, column in METRICS_TABLE_HEADER.items()}


def update_config(config: dict):
    """
    Complete the fields that are not JSON serializable.
    """

    # .dataset.datasets cast all values to Path
    config["dataset"]["datasets"] = {
        name: Path(path) for name, path in config["dataset"]["datasets"].items()
    }

    # .model.encoder.class = FCN_ENCODER
    config["model"]["encoder"]["class"] = FCN_Encoder

    # .model.decoder.class = GlobalHTADecoder
    config["model"]["decoder"]["class"] = GlobalHTADecoder

    # .model.lm.path to Path
    if config["model"].get("lm", {}).get("path"):
        config["model"]["lm"]["path"] = Path(config["model"]["lm"]["path"])

    # Update preprocessing type
    for prepro in config["training"]["data"]["preprocessings"]:
        prepro["type"] = Preprocessing(prepro["type"])

    # .training.output_folder to Path
    config["training"]["output_folder"] = Path(config["training"]["output_folder"])

    if config["training"]["transfer_learning"]:
        # .training.transfer_learning.encoder[1]
        config["training"]["transfer_learning"]["encoder"][1] = Path(
            config["training"]["transfer_learning"]["encoder"][1]
        )

        # .training.transfer_learning.decoder[1]
        config["training"]["transfer_learning"]["decoder"][1] = Path(
            config["training"]["transfer_learning"]["decoder"][1]
        )

    # Parse optimizers
    for optimizer_setup in config["training"]["optimizers"].values():
        # Only supported optimizer is Adam
        optimizer_setup["class"] = Adam

    # set nb_gpu if not present
    if config["training"]["device"]["nb_gpu"] is None:
        config["training"]["device"]["nb_gpu"] = torch.cuda.device_count()


def create_metrics_table(metrics: List[str]) -> PrettyTable:
    """
    Create a Markdown table to display metrics in (CER, WER, NER, etc)
    for each evaluated split.
    """
    table = PrettyTable(
        field_names=["Split"]
        + [title for metric, title in METRICS_TABLE_HEADER.items() if metric in metrics]
    )
    table.set_style(MARKDOWN)

    return table


def add_metrics_table_row(
    table: PrettyTable, split: str, metrics: Optional[Dict[str, int | float]]
) -> list[str]:
    """
    Add a row to an existing metrics Markdown table for the currently evaluated split.
    To create such table please refer to
    [create_metrics_table][dan.ocr.utils.create_metrics_table] function.
    """
    row = [split]
    for column in table.field_names:
        if column not in REVERSE_HEADER:
            continue

        metric_name = REVERSE_HEADER[column]
        metric_value = metrics.get(metric_name)
        row.append(round(metric_value * 100, 2) if metric_value is not None else "−")

    table.add_row(row)
    return row


def load_font(path: Path, size: int):
    """
    Load the font.
    :param path: Path to the font.
    :param size: Size of the font.
    """
    return ImageFont.truetype(path, size)


def search_font_size(
    image: Image,
    font: Path,
    maximum_font_size: int,
    text: str,
    width: int | None = None,
    height: int | None = None,
) -> int:
    """
    Search the biggest font size compatible with the width of the GIF. Take the maximum font size if it is lesser than the perfect font size.
    :param image: Image.
    :param font: Path to the font file.
    :param maximum_font_size: Maximum font size.
    :param text: Predicted text.
    :param width: Image width.
    """
    font_size = maximum_font_size
    font_param = None

    # Check for every font size if it's the perfect font size
    while font_param is None:
        loaded_font = load_font(font, font_size)

        if width is not None:
            # Get place taken by the font
            left, _, right, _ = ImageDraw.Draw(image).multiline_textbbox(
                (width, 0), text, loaded_font
            )

            place_taken = right - left
            font_param = loaded_font if place_taken < width else None

        elif height is not None:
            _, top, _, bottom = ImageDraw.Draw(image).multiline_textbbox(
                (width, 0), text, loaded_font
            )

            place_taken = bottom - top
            font_param = loaded_font if place_taken < round(height / 10) else None

        font_size -= 1

        if font_size == 0:
            logger.warn("No compatible font size found")
            break

    return font_param


def search_spacing(
    image: Image,
    font: FreeTypeFont,
    text: str,
    width: int,
    height: int,
) -> int:
    """
    Search the biggest font size compatible with the width of the GIF. Take the maximum font size if it is lesser than the perfect font size.
    :param image: Image.
    :param font: Parameter of the font.
    :param text: Predicted text.
    :param width: Image width.
    :param height: Image height.
    """
    spacing = 50
    searched_spacing = None

    # Check for every font size if it's the perfect font size
    while searched_spacing is None:
        # Get place taken by the font
        _, _, _, bottom = ImageDraw.Draw(image).multiline_textbbox(
            (width, 0), text, font, spacing=spacing
        )

        searched_spacing = spacing if bottom < height else None

        spacing -= 1

        if spacing == 0:
            logger.warn("No compatible spacing found: font size will be set to 1.")
            searched_spacing = 1

    return searched_spacing


def create_image(
    image: Image,
    text: str,
    font: Path,
    maximum_font_size: int,
    contour: ndarray | None = None,
    scale: float | None = None,
    cer: float | None = None,
    wer: float | None = None,
):
    """
    Create an image with predicted text.
    :param image: Image predicted.
    :param text: Text predicted from the image.
    :param font: Path to the font file.
    :param maximum_font_size: Maximum font size to use.
    :param contour: Contour of the predicted text on the image.
    :param scale: Scaling factor for the output image.
    """
    width, height = image.size

    # Double image size so it have a free with space to write
    new_image = Image.new(image.mode, (width * 2, height), (255, 255, 255))
    new_image.paste(image, (0, 0))
    draw = ImageDraw.Draw(new_image)

    # Search the biggest compatible font size
    font_param = search_font_size(new_image, font, maximum_font_size, text, width)

    if font_param is not None and contour is not None:
        contour = (contour * scale).astype(int32)

        # Get the list of every height of every point of the contour
        heights = [coord[0][1] for coord in contour.tolist()]
        average_height = round(mean(heights))
        draw.text((width, average_height), text, (0, 0, 0), font=font_param)

    elif font_param is not None:
        spacing = search_spacing(new_image, font_param, text, width, height)
        draw.text((width, 0), text, (0, 0, 0), font=font_param, spacing=spacing)

    if cer is None or wer is None:
        return new_image

    more_height = round(height / 10)
    new_height = height + more_height
    cer_wer_text = f"CER : {cer}% | WER : {wer}%"

    # Double image size so it have a free with space to write
    result = Image.new(new_image.mode, (width * 2, new_height), (255, 255, 255))
    result.paste(new_image, (0, more_height))

    draw = ImageDraw.Draw(result)
    font_param = search_font_size(result, font, maximum_font_size, cer_wer_text, height)

    _, _, right, top = draw.textbbox((0, 0), cer_wer_text, font=font_param)

    draw.text(
        ((width * 2 - right) / 2, (more_height - top) / 2),
        cer_wer_text,
        (0, 0, 0),
        font=font_param,
    )

    return result



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
"""
Build all resources needed for the language model from a split extracted by DAN
"""

import pathlib

from dan.datasets.language_model.build import run


def add_language_model_parser(subcommands) -> None:
    parser = subcommands.add_parser(
        "language-model",
        description=__doc__,
        help=__doc__,
    )

    # Required arguments.
    parser.add_argument(
        "--output",
        type=pathlib.Path,
        help="Path where the `labels.json` and `charset.pkl` files are stored and where the data will be generated.",
        required=True,
    )

    # Formatting arguments
    parser.add_argument(
        "--subword-vocab-size",
        type=int,
        help="Size of the vocabulary to train the sentencepiece subword tokenizer needed for language model.",
        default=1000,
    )

    parser.add_argument(
        "--unknown-token",
        type=str,
        default="⁇",
        help="Token to use to replace character in the validation/test sets that is not included in the training set.",
    )
    parser.add_argument(
        "--tokens",
        type=pathlib.Path,
        help="Mapping between starting tokens and end tokens to extract text with their entities.",
        required=False,
    )

    parser.set_defaults(func=run)



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
"""
Each transform class defined here takes as input a PIL Image and returns the modified PIL Image
"""
from enum import Enum
from random import randint

import albumentations as A
import cv2
import numpy as np
from albumentations.augmentations import (
    Affine,
    CoarseDropout,
    ColorJitter,
    ElasticTransform,
    GaussianBlur,
    GaussNoise,
    Perspective,
    RandomScale,
    Sharpen,
    ToGray,
)
from albumentations.core.transforms_interface import ImageOnlyTransform
from cv2 import dilate, erode
from numpy import random
from torch import Tensor
from torchvision.transforms import Compose, ToPILImage
from torchvision.transforms.functional import resize


class Preprocessing(str, Enum):
    MaxResize = "max_resize"
    """
    If the image is bigger than the given size, resize it while keeping the original ratio
    """

    FixedHeightResize = "fixed_height_resize"
    """
    Resize the height to a fixed value while keeping the original ratio
    """

    FixedWidthResize = "fixed_width_resize"
    """
    Resize the width to a fixed value while keeping the original ratio
    """

    FixedResize = "fixed_resize"
    """
    Resize both the width and the height to a fixed value
    """


class FixedHeightResize:
    """
    Resize an image tensor to a fixed height
    """

    def __init__(self, height: int) -> None:
        self.height = height

    def __call__(self, img: Tensor) -> Tensor:
        size = (self.height, self._calc_new_width(img))
        return resize(img, size, antialias=False)

    def _calc_new_width(self, img: Tensor) -> int:
        aspect_ratio = img.shape[2] / img.shape[1]
        return round(self.height * aspect_ratio)


class FixedWidthResize:
    """
    Resize an image tensor to a fixed width
    """

    def __init__(self, width: int) -> None:
        self.width = width

    def __call__(self, img: Tensor) -> Tensor:
        size = (self._calc_new_height(img), self.width)
        return resize(img, size, antialias=False)

    def _calc_new_height(self, img: Tensor) -> int:
        aspect_ratio = img.shape[1] / img.shape[2]
        return round(self.width * aspect_ratio)


class FixedResize:
    """
    Resize an image tensor to a fixed width and height
    """

    def __init__(self, height: int, width: int) -> None:
        self.height = height
        self.width = width

    def __call__(self, img: Tensor) -> Tensor:
        return resize(img, (self.height, self.width), antialias=False)


class MaxResize:
    """
    Resize an image tensor if it is bigger than the maximum size
    """

    def __init__(self, height: int, width: int) -> None:
        self.max_width = width
        self.max_height = height

    def __call__(self, img: Tensor) -> Tensor:
        height, width = img.shape[1:]
        if width <= self.max_width and height <= self.max_height:
            return img
        width_ratio = self.max_width / width
        height_ratio = self.max_height / height
        ratio = min(height_ratio, width_ratio)
        new_width = int(width * ratio)
        new_height = int(height * ratio)
        return resize(img, (new_height, new_width), antialias=False)


class Dilation:
    """
    OCR: stroke width increasing
    """

    def __init__(self, kernel, iterations):
        self.kernel = kernel
        self.iterations = iterations

    def __call__(self, x):
        return dilate(np.array(x), self.kernel, iterations=self.iterations)


class Erosion:
    """
    OCR: stroke width decreasing
    """

    def __init__(self, kernel, iterations):
        self.kernel = kernel
        self.iterations = iterations

    def __call__(self, x):
        return erode(np.array(x), self.kernel, iterations=self.iterations)


class ErosionDilation(ImageOnlyTransform):
    """
    Random erosion or dilation
    """

    def __init__(
        self,
        min_kernel: int,
        max_kernel: int,
        iterations: int,
        always_apply: bool = False,
        p: float = 1.0,
    ):
        super(ErosionDilation, self).__init__(always_apply, p)
        self.min_kernel = min_kernel
        self.max_kernel = max_kernel
        self.iterations = iterations
        self.p = p
        self.always_apply = False

    def apply(self, img: np.ndarray, **params):
        kernel_h = randint(self.min_kernel, self.max_kernel)
        kernel_w = randint(self.min_kernel, self.max_kernel)
        kernel = np.ones((kernel_h, kernel_w), np.uint8)
        return (
            Erosion(kernel, iterations=self.iterations)(img)
            if random.random() < 0.5
            else Dilation(kernel=kernel, iterations=self.iterations)(img)
        )


def get_preprocessing_transforms(
    preprocessings: list, to_pil_image: bool = False
) -> Compose:
    """
    Returns a list of transformations to be applied to the image.
    """
    transforms = []
    for preprocessing in preprocessings:
        match preprocessing["type"]:
            case Preprocessing.MaxResize:
                transforms.append(
                    MaxResize(
                        height=preprocessing["max_height"],
                        width=preprocessing["max_width"],
                    )
                )
            case Preprocessing.FixedHeightResize:
                transforms.append(
                    FixedHeightResize(height=preprocessing["fixed_height"])
                )
            case Preprocessing.FixedWidthResize:
                transforms.append(FixedWidthResize(width=preprocessing["fixed_width"]))
            case Preprocessing.FixedResize:
                transforms.append(
                    FixedResize(
                        height=preprocessing["fixed_height"],
                        width=preprocessing["fixed_width"],
                    )
                )
    if to_pil_image:
        transforms.append(ToPILImage())
    return Compose(transforms)


def get_augmentation_transforms() -> A.Compose:
    """
    Returns a list of transformation to be applied to the image.
    """
    return A.Compose(
        [
            # Scale between 0.75 and 1.0
            RandomScale(
                scale_limit=[-0.25, 0], always_apply=True, interpolation=cv2.INTER_AREA
            ),
            A.SomeOf(
                [
                    ErosionDilation(min_kernel=1, max_kernel=4, iterations=1),
                    Perspective(scale=(0.05, 0.09), fit_output=True, p=0.4),
                    GaussianBlur(sigma_limit=2.5, p=1),
                    GaussNoise(var_limit=50**2, p=1),
                    ColorJitter(
                        contrast=0.2,
                        brightness=0.2,
                        saturation=0.2,
                        hue=0.2,
                        p=1,
                    ),
                    ElasticTransform(
                        alpha=20.0,
                        sigma=5.0,
                        alpha_affine=1.0,
                        border_mode=0,
                        p=1,
                    ),
                    Sharpen(alpha=(0.0, 1.0), p=1),
                    Affine(shear={"x": (-20, 20), "y": (0, 0)}, p=1),
                    CoarseDropout(p=1),
                    ToGray(p=0.5),
                ],
                n=2,
                p=0.9,
            ),
        ],
        p=0.9,
    )



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import functools
import logging
import os
from contextlib import contextmanager

import requests

logger = logging.getLogger(__name__)

try:
    import mlflow
    from mlflow.environment_variables import MLFLOW_HTTP_REQUEST_MAX_RETRIES

    MLFLOW_AVAILABLE = True
    logger.info("MLflow logging is available.")
except ImportError:
    MLFLOW_AVAILABLE = False


def mlflow_required(func):
    """
    Always check that MLflow is available before executing the function.
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if not MLFLOW_AVAILABLE:
            return
        return func(self, *args, **kwargs)

    return wrapper


@mlflow_required
def make_mlflow_request(mlflow_method, *args, **kwargs):
    """
    Encapsulate MLflow HTTP requests to prevent them from crashing the whole training process.
    """
    try:
        mlflow_method(*args, **kwargs)
    except requests.exceptions.ConnectionError as e:
        logger.error(f"Call to `{str(mlflow_method)}` failed with error: {str(e)}")


@mlflow_required
def setup_environment(config: dict):
    """
    Get the necessary variables from the config file and put them in the environment variables

    :param config: dict, the config of the model
    """
    needed_variables = {
        "MLFLOW_S3_ENDPOINT_URL": "s3_endpoint_url",
        "MLFLOW_TRACKING_URI": "tracking_uri",
        "AWS_ACCESS_KEY_ID": "aws_access_key_id",
        "AWS_SECRET_ACCESS_KEY": "aws_secret_access_key",
    }
    for variable_name, config_key in needed_variables.items():
        if config_key in config:
            os.environ[variable_name] = config[config_key]

    # Check max retry setting
    max_retries = MLFLOW_HTTP_REQUEST_MAX_RETRIES.get()
    if max_retries and int(max_retries) <= 1:
        logger.warning(
            f"The maximum number of retries for MLflow HTTP requests is set to {max_retries}, which is low. Consider using a higher value."
        )


@mlflow_required
def logging_metrics(
    display_values: dict,
    step: str,
    epoch: int,
    mlflow_logging: bool = False,
    is_master: bool = False,
):
    """
    Log dictionary metrics in the Metrics section of MLflow

    :param display_values: dict, the dictionary containing the metrics to publish on MLflow
    :param step: str, the step for which the metrics are to be published on Metrics section (ex: train, val, test). This will allow a better display on MLflow.
    :param epoch: int, the current epoch.
    :param mlflow_logging: bool, allows you to verify that you have the authorization to log on MLflow, defaults to False
    :param is_master: bool, makes sure you're on the right thread, defaults to False
    """
    if mlflow_logging and is_master:
        make_mlflow_request(
            mlflow_method=mlflow.log_metrics,
            metrics={f"{step}_{name}": value for name, value in display_values.items()},
            step=epoch,
        )


@mlflow_required
def logging_tags_metrics(
    display_values: dict,
    step: str,
    mlflow_logging: bool = False,
    is_master: bool = False,
):
    """
    Log dictionary metrics in the Tags section of MLflow

    :param display_values: dict, the dictionary containing the metrics to publish on MLflow
    :param step: str, the step for which the metrics are to be published on Tags section (ex: train, val, test). This will allow a better display on MLflow.
    :param mlflow_logging: bool, allows you to verify that you have the authorization to log on MLflow, defaults to False
    :param is_master: bool, makes sure you're on the right thread, defaults to False
    """
    if mlflow_logging and is_master:
        make_mlflow_request(
            mlflow_method=mlflow.set_tags,
            tags={f"{step}_{name}": value for name, value in display_values.items()},
        )


@mlflow_required
@contextmanager
def start_mlflow_run(config: dict):
    """
    Create an MLflow execution context with the parameters contained in the config file.

    Yields the active MLflow run, as well as a boolean saying whether a new one was created.

    :param config: dict, the config of the model
    """

    # Set needed variables in environment
    setup_environment(config)

    run_name, run_id = config.get("run_name"), config.get("run_id")

    if run_id:
        logger.info(f"Will resume run ({run_id}).")

        if run_name:
            logger.warning(
                "Run_name will be ignored since you specified a run_id to resume from."
            )

    # Set experiment from config
    experiment_id = config.get("experiment_id")
    assert experiment_id, "Missing MLflow experiment ID in the configuration"

    # Start run
    yield (
        mlflow.start_run(run_id=run_id, run_name=run_name, experiment_id=experiment_id),
        run_id is None,
    )
    mlflow.end_run()



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

import shutil

import pytest

from dan.ocr.predict.attention import (
    Level,
    parse_delimiters,
    split_text,
    split_text_and_confidences,
)
from dan.ocr.predict.inference import run as run_prediction
from dan.utils import (
    EntityType,
    parse_charset_pattern,
    parse_tokens,
    parse_tokens_pattern,
)
from tests import FIXTURES

PREDICTION_DATA_PATH = FIXTURES / "prediction"

FONT_PATH = FIXTURES.parent.parent / "fonts/LinuxLibertine.ttf"


@pytest.mark.parametrize(
    (
        "text",
        "confidence",
        "level",
        "tokens",
        "expected_split_text",
        "expected_mean_confidences",
        "expected_offsets",
    ),
    [
        # level: char
        (
            "To <kyo>",
            [0.1, 0.2, 0.3, 0.4],
            Level.Char,
            None,
            ["T", "o", " ", "<kyo>"],
            [0.1, 0.2, 0.3, 0.4],
            [0, 0, 0, 0],
        ),
        # level: word
        (
            "Lo ve\nTokyo",
            [0.1, 0.1, 0.2, 0.3, 0.3, 0.4, 0.5, 0.5, 0.5, 0.5, 0.5],
            Level.Word,
            None,
            ["Lo", "ve", "Tokyo"],
            [0.1, 0.3, 0.5],
            [1, 1, 0],
        ),
        # level: line
        (
            "Love\nTokyo",
            [0.1, 0.1, 0.1, 0.1, 0.2, 0.3, 0.3, 0.3, 0.3, 0.3, 0.3],
            Level.Line,
            None,
            ["Love", "Tokyo"],
            [0.1, 0.3],
            [1, 0],
        ),
        # level: NER (no end tokens)
        (
            "ⒶLove ⒷTokyo",
            [0.1, 0.2, 0.2, 0.2, 0.2, 0.3, 0.4, 0.5, 0.5, 0.5, 0.5, 0.5],
            Level.NER,
            [EntityType(start="Ⓐ"), EntityType(start="Ⓑ")],
            ["ⒶLove ", "ⒷTokyo"],
            [0.2, 0.48],
            [0, 0],
        ),
        # level: NER (with end tokens)
        (
            "ⓐLoveⒶ ⓑTokyoⒷ",
            [0.1, 0.2, 0.2, 0.2, 0.2, 0.3, 0.4, 0.5, 0.6, 0.6, 0.6, 0.6, 0.6, 0.7],
            Level.NER,
            [EntityType(start="ⓐ", end="Ⓐ"), EntityType(start="ⓑ", end="Ⓑ")],
            ["ⓐLoveⒶ", "ⓑTokyoⒷ"],
            [0.2, 0.6],
            [1, 0],
        ),
        # level: NER (no end tokens, no space)
        (
            "ⒶLoveⒷTokyo",
            [0.1, 0.2, 0.2, 0.2, 0.2, 0.3, 0.4, 0.4, 0.4, 0.4, 0.4],
            Level.NER,
            [EntityType(start="Ⓐ"), EntityType(start="Ⓑ")],
            ["ⒶLove", "ⒷTokyo"],
            [0.18, 0.38],
            [0, 0],
        ),
        # level: NER (with end tokens, no space)
        (
            "ⓐLoveⒶⓑTokyoⒷ",
            [0.1, 0.2, 0.2, 0.2, 0.2, 0.3, 0.4, 0.5, 0.5, 0.5, 0.5, 0.5, 0.6],
            Level.NER,
            [EntityType(start="ⓐ", end="Ⓐ"), EntityType(start="ⓑ", end="Ⓑ")],
            ["ⓐLoveⒶ", "ⓑTokyoⒷ"],
            [0.2, 0.5],
            [0, 0],
        ),
    ],
)
def test_split_text_and_confidences(
    text: str,
    confidence: list[float],
    level: Level,
    tokens: list[EntityType] | None,
    expected_split_text: list[str],
    expected_mean_confidences: list[list[float]],
    expected_offsets: list[int],
):
    # Full charset
    charset = [
        # alphabet
        "T",
        "o",
        "L",
        "v",
        "e",
        "k",
        "y",
        # Entities
        "ⓐ",
        "Ⓐ",
        "ⓑ",
        "Ⓑ",
        # Special
        "<kyo>",
        # Punctuation
        " ",
    ]
    texts_conf, averages_conf, offsets_conf = split_text_and_confidences(
        text=text,
        confidences=confidence,
        level=level,
        char_separators=parse_charset_pattern(charset),
        word_separators=parse_delimiters([" ", "\n"]),
        line_separators=parse_delimiters(["\n"]),
        tokens_separators=parse_tokens_pattern(tokens) if tokens else None,
    )
    texts, offsets = split_text(
        text=text,
        level=level,
        char_separators=parse_charset_pattern(charset),
        word_separators=parse_delimiters([" ", "\n"]),
        line_separators=parse_delimiters(["\n"]),
        tokens_separators=parse_tokens_pattern(tokens) if tokens else None,
    )

    assert texts == expected_split_text
    assert offsets == expected_offsets
    assert texts_conf == expected_split_text
    assert averages_conf == expected_mean_confidences
    assert offsets_conf == expected_offsets


def test_plot_attention(
    tmp_path,
):
    confidence_score = []

    image_name = "0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84"

    image_dir = tmp_path / "images"
    image_dir.mkdir()
    shutil.copyfile(
        (PREDICTION_DATA_PATH / "images" / image_name).with_suffix(".png"),
        (image_dir / image_name).with_suffix(".png"),
    )

    run_prediction(
        image_dir=image_dir,
        font=FONT_PATH,
        maximum_font_size=32,
        model=PREDICTION_DATA_PATH,
        output=tmp_path,
        confidence_score=bool(confidence_score),
        confidence_score_levels=confidence_score,
        attention_map=True,
        attention_map_level=[Level.Line].pop(),
        attention_map_scale=0.5,
        alpha_factor=0.9,
        color_map="nipy_spectral",
        word_separators=[" ", "\n"],
        line_separators=["\n"],
        temperature=1.0,
        predict_objects=True,
        max_object_height=None,
        image_extension=".png",
        gpu_device=None,
        batch_size=1,
        tokens=parse_tokens(PREDICTION_DATA_PATH / "tokens.yml"),
        start_token=None,
        use_language_model=False,
        compile_model=False,
        dynamic_mode=False,
    )

    outname = image_name + "_line.gif"

    assert (tmp_path / outname).read_bytes() == (
        PREDICTION_DATA_PATH / outname
    ).read_bytes()



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
"""
Generate the YAML file containing entities and their token(s) to train a DAN model
"""
from pathlib import Path

from dan.datasets.tokens.generate import run


def add_tokens_parser(subcommands) -> None:
    parser = subcommands.add_parser(
        "tokens",
        description=__doc__,
        help=__doc__,
    )
    parser.add_argument(
        "entities",
        type=Path,
        help="Path to a YAML file containing the extracted entities.",
    )
    parser.add_argument(
        "--end-tokens",
        action="store_true",
        help="Whether to generate end tokens along with starting tokens.",
    )
    parser.add_argument(
        "--output-file",
        type=Path,
        default=Path("tokens.yml"),
        required=False,
        help="Path to a YAML file to save the entities and their token(s).",
    )

    parser.set_defaults(func=run)



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import json
import logging
import pickle
from operator import attrgetter, methodcaller
from pathlib import Path

import pytest
from PIL import Image, ImageChops

from dan.datasets.download.images import IIIF_FULL_SIZE, ImageDownloader
from dan.datasets.download.utils import download_image
from dan.utils import parse_tokens
from line_image_extractor.image_utils import BoundingBox
from tests import FIXTURES, change_split_content

EXTRACTION_DATA_PATH = FIXTURES / "extraction"


@pytest.mark.parametrize(
    "max_width, max_height, width, height, resize",
    (
        (1000, 2000, 900, 800, IIIF_FULL_SIZE),
        (1000, 2000, 1100, 800, "1000,"),
        (1000, 2000, 1100, 2800, ",2000"),
        (1000, 2000, 2000, 3000, "1000,"),
    ),
)
def test_get_iiif_size_arg(max_width, max_height, width, height, resize, tmp_path):
    split_path = tmp_path / "output" / "split.json"
    split_path.parent.mkdir()
    split_path.write_text(json.dumps({"train": {}}))

    assert (
        ImageDownloader(
            output=split_path.parent, max_width=max_width, max_height=max_height
        ).get_iiif_size_arg(width=width, height=height)
        == resize
    )


@pytest.mark.parametrize("load_entities", [True, False])
@pytest.mark.parametrize("keep_spaces", [True, False])
@pytest.mark.parametrize(
    "transcription_entities_worker_version", ["worker_version_id", False]
)
def test_download(
    load_entities,
    keep_spaces,
    transcription_entities_worker_version,
    split_content,
    monkeypatch,
    tmp_path,
):
    output = tmp_path / "download"
    output.mkdir(parents=True, exist_ok=True)

    # Mock tokens
    tokens_path = EXTRACTION_DATA_PATH / "tokens.yml"
    tokens = [
        token
        for entity_type in parse_tokens(tokens_path).values()
        for token in [entity_type.start, entity_type.end]
        if token
    ]

    # Mock "split.json"
    split_content, expected_labels = change_split_content(
        load_entities,
        transcription_entities_worker_version,
        keep_spaces,
        split_content,
        tokens,
        {
            "test": {
                "images/test/dataset_id/test-page_1-line_1.jpg": "ⓢLeunaut  ⓕClau⁇e  ⓑ⁇⁇",
                "images/test/dataset_id/test-page_1-line_2.jpg": "ⓢ⁇aurac⁇o  ⓕClau⁇ine  ⓑ⁇⁇",
                "images/test/dataset_id/test-page_1-line_3.jpg": "ⓢLaurent  ⓕJac⁇use  ⓑ21",
                "images/test/dataset_id/test-page_2-line_1.jpg": "ⓢ⁇alette  ⓕElisa⁇et⁇  ⓑ7⁇",
                "images/test/dataset_id/test-page_2-line_2.jpg": "ⓢTan⁇ol  ⓕJean  ⓑ7⁇",
                "images/test/dataset_id/test-page_2-line_3.jpg": "ⓢ⁇auret  ⓕJean  ⓑ⁇⁇",
            },
            "train": {
                "images/train/dataset_id/train-page_1-line_1.jpg": "ⓢLaulont  ⓕFrancois  ⓑ8",
                "images/train/dataset_id/train-page_1-line_2.jpg": "ⓢCiret  ⓕAntoine  ⓑ27",
                "images/train/dataset_id/train-page_1-line_3.jpg": "ⓢCiret  ⓕMarie  ⓑ28",
                "images/train/dataset_id/train-page_1-line_4.jpg": "ⓢCiret  ⓕMarie  ⓑ2",
                "images/train/dataset_id/train-page_2-line_1.jpg": "ⓢEureston  ⓕSolange  ⓑ10",
                "images/train/dataset_id/train-page_2-line_2.jpg": "ⓢTerontussieux  ⓕJean  ⓑ2",
                "images/train/dataset_id/train-page_2-line_3.jpg": "ⓢPressonet  ⓕMarie  ⓑ12",
            },
            "val": {
                "images/val/dataset_id/val-page_1-line_1.jpg": "ⓢCirau⁇  ⓕAntoine  ⓑ⁇⁇",
                "images/val/dataset_id/val-page_1-line_2.jpg": "ⓢCirau⁇  ⓕPriser  ⓑ⁇⁇",
                "images/val/dataset_id/val-page_1-line_3.jpg": "ⓢCirau⁇  ⓕElisa⁇et⁇  ⓑ⁇⁇",
            },
        },
    )
    (output / "split.json").write_text(json.dumps(split_content))

    # Mock download_image so that it simply opens it with Pillow
    monkeypatch.setattr(
        "dan.datasets.download.images.download_image", lambda url: Image.open(url)
    )

    def mock_build_image_url(polygon, image_url, *args, **kwargs):
        # During tests, the image URL is its local path
        return image_url

    extractor = ImageDownloader(
        output=output,
        image_extension=".jpg",
    )
    # Mock build_image_url to simply return the path to the image
    extractor.build_iiif_url = mock_build_image_url
    extractor.run()

    # Check files
    IMAGE_DIR = output / "images"
    TEST_DIR = IMAGE_DIR / "test" / "dataset_id"
    TRAIN_DIR = IMAGE_DIR / "train" / "dataset_id"
    VAL_DIR = IMAGE_DIR / "val" / "dataset_id"

    expected_paths = [
        output / "charset.pkl",
        # Images of test folder
        TEST_DIR / "test-page_1-line_1.jpg",
        TEST_DIR / "test-page_1-line_2.jpg",
        TEST_DIR / "test-page_1-line_3.jpg",
        TEST_DIR / "test-page_2-line_1.jpg",
        TEST_DIR / "test-page_2-line_2.jpg",
        TEST_DIR / "test-page_2-line_3.jpg",
        # Images of train folder
        TRAIN_DIR / "train-page_1-line_1.jpg",
        TRAIN_DIR / "train-page_1-line_2.jpg",
        TRAIN_DIR / "train-page_1-line_3.jpg",
        TRAIN_DIR / "train-page_1-line_4.jpg",
        TRAIN_DIR / "train-page_2-line_1.jpg",
        TRAIN_DIR / "train-page_2-line_2.jpg",
        TRAIN_DIR / "train-page_2-line_3.jpg",
        # Images of val folder
        VAL_DIR / "val-page_1-line_1.jpg",
        VAL_DIR / "val-page_1-line_2.jpg",
        VAL_DIR / "val-page_1-line_3.jpg",
        output / "labels.json",
        output / "split.json",
    ]
    assert sorted(filter(methodcaller("is_file"), output.rglob("*"))) == expected_paths

    # Check "charset.pkl"
    expected_charset = {"⁇"}
    for values in split_content["train"].values():
        expected_charset.update(set(values["text"]))

    if load_entities:
        expected_charset.update(tokens)

    assert set(pickle.loads((output / "charset.pkl").read_bytes())) == expected_charset

    # Check "labels.json"
    assert json.loads((output / "labels.json").read_text()) == expected_labels

    # Check cropped images
    for expected_path in expected_paths:
        if expected_path.suffix != ".jpg":
            continue

        assert ImageChops.difference(
            Image.open(EXTRACTION_DATA_PATH / "images" / expected_path.name),
            Image.open(expected_path),
        )


def test_download_image_error(monkeypatch, caplog, capsys, tmp_path):
    task = {
        "split": "train",
        "polygon": [],
        "image_url": "deadbeef",
        "destination": Path("/dev/null"),
    }
    monkeypatch.setattr(
        "dan.datasets.download.images.polygon_to_bbox",
        lambda polygon: BoundingBox(0, 0, 0, 0),
    )

    split_path = tmp_path / "output" / "split.json"
    split_path.parent.mkdir()
    split_path.write_text(json.dumps({"train": {}}))

    extractor = ImageDownloader(output=split_path.parent, image_extension=".jpg")

    # Add the key in data
    extractor.data[task["split"]][str(task["destination"])] = "deadbeefdata"

    # Build a random task
    extractor.download_images([task])

    # Key should have been removed
    assert str(task["destination"]) not in extractor.data[task["split"]]

    # Check error log
    assert len(caplog.record_tuples) == 1
    _, level, msg = caplog.record_tuples[0]
    assert level == logging.ERROR
    assert msg == "Failed to download 1 image(s)."

    # Check stdout
    captured = capsys.readouterr()
    assert captured.out == "deadbeef: Image URL must be HTTP(S) for element null\n"


def test_download_image_error_try_max(responses, caplog):
    # An image's URL
    url = (
        "https://blabla.com/iiif/2/image_path.jpg/231,699,2789,3659/full/0/default.jpg"
    )
    fixed_url = (
        "https://blabla.com/iiif/2/image_path.jpg/231,699,2789,3659/max/0/default.jpg"
    )

    # Fake responses error
    responses.add(
        responses.GET,
        url,
        status=400,
    )
    # Correct response with max
    responses.add(
        responses.GET,
        fixed_url,
        status=200,
        body=next((FIXTURES / "prediction" / "images").iterdir()).read_bytes(),
    )

    image = download_image(url)

    assert image
    # We try 3 times with the first URL
    # Then the first try with the new URL is successful
    assert len(responses.calls) == 4
    assert list(map(attrgetter("request.url"), responses.calls)) == [url] * 3 + [
        fixed_url
    ]

    # Check error log
    assert len(caplog.record_tuples) == 2

    # We should only have WARNING levels
    assert set(level for _, level, _ in caplog.record_tuples) == {logging.WARNING}



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
"""
Evaluate a trained DAN model.
"""

import json
import logging
import random
from argparse import ArgumentTypeError
from itertools import chain
from operator import attrgetter
from pathlib import Path
from typing import Dict, List

import numpy as np
import torch
import torch.multiprocessing as mp
from edlib import align, getNiceAlignment
from nerval.evaluate import evaluate
from nerval.parse import parse_bio
from nerval.utils import TABLE_HEADER as NERVAL_TABLE_HEADER
from nerval.utils import print_results
from prettytable import MARKDOWN, PrettyTable

from dan.bio import convert
from dan.ocr import wandb
from dan.ocr.manager.metrics import Inference
from dan.ocr.manager.training import Manager
from dan.ocr.utils import add_metrics_table_row, create_metrics_table, update_config
from dan.utils import parse_tokens, read_json

logger = logging.getLogger(__name__)

NERVAL_THRESHOLD = 0.30
NB_WORST_PREDICTIONS = 5

EMPTY_STRING = "∅"


def parse_threshold(value: str) -> float:
    """
    Check that the string passed as parameter is a correct floating point number between 0 and 1
    """
    try:
        value = float(value)
    except ValueError:
        raise ArgumentTypeError("Must be a floating point number.")

    if value < 0 or value > 1:
        raise ArgumentTypeError("Must be between 0 and 1.")

    return value


def add_evaluate_parser(subcommands) -> None:
    parser = subcommands.add_parser(
        "evaluate",
        description=__doc__,
        help=__doc__,
    )

    parser.add_argument(
        "--config",
        type=read_json,
        required=True,
        help="Configuration file.",
    )

    parser.add_argument(
        "--nerval-threshold",
        help="Distance threshold for the match between gold and predicted entity during Nerval evaluation.",
        default=NERVAL_THRESHOLD,
        type=parse_threshold,
    )

    parser.add_argument(
        "--output-json",
        help="Where to save evaluation results in JSON format.",
        default=None,
        type=Path,
    )

    parser.set_defaults(func=run)


def print_worst_predictions(all_inferences: Dict[str, List[Inference]]):
    table_header, table_values = (
        [
            "Image name",
            "WER",
            "Alignment between ground truth - prediction",
        ],
        [],
    )

    worst_inferences = sorted(
        chain.from_iterable(all_inferences.values()),
        key=attrgetter("wer"),
        reverse=True,
    )[:NB_WORST_PREDICTIONS]
    for inference in worst_inferences:
        if not inference.ground_truth:
            logger.warning(
                f"Ground truth is empty for {inference.image}. `{EMPTY_STRING}` will be displayed"
            )

        if not inference.prediction:
            logger.warning(
                f"Prediction is empty for {inference.image}. `{EMPTY_STRING}` will be displayed"
            )

        alignment = getNiceAlignment(
            align(
                inference.ground_truth or EMPTY_STRING,
                inference.prediction or EMPTY_STRING,
                task="path",
            ),
            inference.ground_truth or EMPTY_STRING,
            inference.prediction or EMPTY_STRING,
        )
        alignment_str = f'{alignment["query_aligned"]}\n{alignment["matched_aligned"]}\n{alignment["target_aligned"]}'
        table_values.append(
            [inference.image, round(inference.wer * 100, 2), alignment_str]
        )

    # Display/Log tables
    print(f"\n#### {NB_WORST_PREDICTIONS} worst prediction(s)\n")

    table = PrettyTable(field_names=table_header)
    table.set_style(MARKDOWN)
    table.add_rows(table_values)
    print(table)

    wandb.table(
        f"evaluation/{NB_WORST_PREDICTIONS}_worst_predictions",
        columns=table_header,
        data=table_values,
    )


def eval_nerval(
    all_inferences: Dict[str, List[Inference]],
    tokens: Path,
    threshold: float,
):
    print("\n#### Nerval evaluation")

    def inferences_to_parsed_bio(attr: str):
        bio_values = []
        for inference in inferences:
            value = getattr(inference, attr)
            bio_value = convert(value, ner_tokens=tokens)
            bio_values.extend(bio_value.split("\n"))

        # Parse this BIO format
        return parse_bio(bio_values)

    # Evaluate with Nerval
    tokens = parse_tokens(tokens)
    for split_name, inferences in all_inferences.items():
        ground_truths = inferences_to_parsed_bio("ground_truth")
        predictions = inferences_to_parsed_bio("prediction")

        if not (ground_truths and predictions):
            continue

        scores = {
            key: {
                k: round(value * 100, 2)
                if k in ["P", "R", "F1"] and value is not None
                # Value can be None when
                # no entity is predicted for a specific entity type
                # there is no entity in the labels for a specific entity type
                else (value or 0)
                for k, value in values.items()
            }
            for key, values in evaluate(ground_truths, predictions, threshold).items()
        }

        # Display/Log tables
        print(f"\n##### {split_name}\n")
        metrics_values = print_results(scores)
        wandb.table(
            f"evaluation/nerval/{split_name}",
            columns=NERVAL_TABLE_HEADER,
            data=metrics_values,
        )


def eval(
    rank,
    config: dict,
    nerval_threshold: float,
    output_json: Path | None,
    mlflow_logging: bool,
):
    torch.manual_seed(0)
    torch.cuda.manual_seed(0)
    np.random.seed(0)
    random.seed(0)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True

    config["training"]["device"]["ddp_rank"] = rank

    # Load best checkpoint
    config["training"]["load_epoch"] = "best"

    model = Manager(config)
    model.load_model()

    metric_names = [
        "cer",
        "cer_no_token",
        "wer",
        "wer_no_punct",
        "wer_no_token",
        "time",
    ]
    if config["dataset"]["tokens"] is not None:
        metric_names.append("ner")

    metrics_table = create_metrics_table(metric_names)
    metrics_values, all_inferences = [], {}

    for dataset_name in config["dataset"]["datasets"]:
        for set_name in ["train", "val", "test"]:
            logger.info(f"Evaluating on set `{set_name}`")
            metrics, inferences = model.evaluate(
                "{}-{}".format(dataset_name, set_name),
                [
                    (dataset_name, set_name),
                ],
                metric_names,
                mlflow_logging=mlflow_logging,
            )

            metrics_values.append(
                add_metrics_table_row(metrics_table, set_name, metrics)
            )
            all_inferences[set_name] = inferences

    # Display/Log tables
    print("\n#### DAN evaluation\n")
    print(metrics_table)
    wandb.table(
        "evaluation/dan", columns=metrics_table.field_names, data=metrics_values
    )

    if "ner" in metric_names:
        eval_nerval(
            all_inferences,
            tokens=config["dataset"]["tokens"],
            threshold=nerval_threshold,
        )

    print_worst_predictions(all_inferences)

    # Save to JSON
    if output_json is not None:
        output_json.write_text(json.dumps(all_inferences, indent=2))

        if config.get("wandb", {}).get("inferences"):
            artifact = wandb.artifact(
                name=f"run-{wandb.run_id()}-evaluation",
                type="json",
                description="Evaluation metrics",
            )
            wandb.log_artifact(artifact, local_path=output_json, name=output_json.name)


def run(config: dict, nerval_threshold: float, output_json: Path | None):
    update_config(config)

    # Start "Weights & Biases" as soon as possible
    wandb.init(config)

    mlflow_logging = bool(config.get("mlflow"))

    if mlflow_logging:
        logger.info("MLflow logging enabled")

    if (
        config["training"]["device"]["use_ddp"]
        and config["training"]["device"]["force"] in [None, "cuda"]
        and torch.cuda.is_available()
    ):
        mp.spawn(
            eval,
            args=(config, nerval_threshold, output_json, mlflow_logging),
            nprocs=config["training"]["device"]["nb_gpu"],
        )
    else:
        eval(0, config, nerval_threshold, output_json, mlflow_logging)



import functools
import os
from pathlib import Path
from typing import TYPE_CHECKING

import wandb

if TYPE_CHECKING:
    from wandb.sdk.data_types.image import ImageDataOrPathType


WANDB_AVAILABLE = False


def wandb_required(func):
    """
    Always check that "Weights & Biases" is available before executing the function.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not WANDB_AVAILABLE:
            return
        return func(*args, **kwargs)

    return wrapper


def init(config: dict) -> None:
    if not config.get("wandb"):
        return

    # Store "Weights & Biases" files in the output folder by default
    if not os.getenv("WANDB_DIR"):
        os.environ["WANDB_DIR"] = str(config["training"]["output_folder"])
    # Path should already exist when "Weights & Biases" is instantiated
    Path(os.environ["WANDB_DIR"]).mkdir(parents=True, exist_ok=True)

    wandb_params = config["wandb"].get("init", {})
    wandb_config = wandb_params.pop("config", {})
    wandb.init(
        **wandb_params,
        config={**config, **wandb_config},
    )

    global WANDB_AVAILABLE
    WANDB_AVAILABLE = True


@wandb_required
def log(*args, **kwargs) -> None:
    wandb.log(*args, **kwargs)


@wandb_required
def image(image_name: str, image_data: "ImageDataOrPathType", *args, **kwargs) -> None:
    image = wandb.Image(image_data)
    log(data={image_name: image}, *args, **kwargs)


@wandb_required
def table(table_name: str, *args, **kwargs) -> None:
    table = wandb.Table(*args, **kwargs)
    log(data={table_name: table}, commit=True)


@wandb_required
def artifact(*args, **kwargs) -> wandb.Artifact | None:
    return wandb.Artifact(*args, **kwargs)


@wandb_required
def log_artifact(artifact: wandb.Artifact, *args, **kwargs) -> None:
    artifact.add_file(*args, **kwargs)
    wandb.log_artifact(artifact)


@wandb_required
def run_id() -> str | None:
    if not wandb.run:
        return

    return wandb.run.id



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

import json
import re
from operator import methodcaller
from typing import NamedTuple

import pytest

from arkindex_export import (
    Element,
    Transcription,
    TranscriptionEntity,
)
from dan.datasets.extract.arkindex import ArkindexExtractor
from dan.datasets.extract.db import get_transcription_entities
from dan.datasets.extract.exceptions import (
    NoTranscriptionError,
)
from dan.datasets.extract.utils import (
    EntityType,
    entities_to_xml,
    normalize_linebreaks,
    normalize_spaces,
)
from dan.utils import parse_tokens
from tests import FIXTURES, change_split_content

EXTRACTION_DATA_PATH = FIXTURES / "extraction"

ENTITY_TOKEN_SPACE = re.compile(r"[ⓢ|ⓕ|ⓑ] ")
TWO_SPACES_LM_REGEX = re.compile(r"▁ ▁")

# NamedTuple to mock actual database result
Entity = NamedTuple("Entity", offset=int, length=int, type=str, value=str)

TOKENS = {
    "P": EntityType(start="ⓟ", end="Ⓟ"),
    "D": EntityType(start="ⓓ", end="Ⓓ"),
    "N": EntityType(start="ⓝ", end="Ⓝ"),
    "I": EntityType(start="ⓘ", end="Ⓘ"),
}


def filter_tokens(keys):
    return {key: value for key, value in TOKENS.items() if key in keys}


@pytest.mark.parametrize(
    "text,trimmed",
    (
        ("no_spaces", "no_spaces"),
        (" beginning", "beginning"),
        ("ending ", "ending"),
        (" both ", "both"),
        ("    consecutive", "consecutive"),
        ("\ttab", "tab"),
        ("\t tab", "tab"),
        (" \ttab", "tab"),
        ("no|space", "no|space"),
    ),
)
def test_normalize_spaces(text, trimmed):
    assert normalize_spaces(text) == trimmed


@pytest.mark.parametrize(
    "text,trimmed",
    (
        ("no_linebreaks", "no_linebreaks"),
        ("\nbeginning", "beginning"),
        ("ending\n", "ending"),
        ("\nboth\n", "both"),
        ("\n\n\nconsecutive", "consecutive"),
        ("\rcarriage_return", "carriage_return"),
        ("\r\ncarriage_return+linebreak", "carriage_return+linebreak"),
        ("\n\r\r\n\ncarriage_return+linebreak", "carriage_return+linebreak"),
        ("no|linebreaks", "no|linebreaks"),
    ),
)
def test_normalize_linebreaks(text, trimmed):
    assert normalize_linebreaks(text) == trimmed


@pytest.mark.parametrize("load_entities", [True, False])
@pytest.mark.parametrize("keep_spaces", [True, False])
@pytest.mark.parametrize(
    "transcription_entities_worker_version", ["worker_version_id", False]
)
@pytest.mark.parametrize("existing", ((True, False)))
def test_extract(
    load_entities,
    keep_spaces,
    transcription_entities_worker_version,
    split_content,
    mock_database,
    tmp_path,
    existing,
):
    output = tmp_path / "extraction"
    output.mkdir(parents=True, exist_ok=True)

    # Mock tokens
    tokens_path = EXTRACTION_DATA_PATH / "tokens.yml"
    tokens = [
        token
        for entity_type in parse_tokens(tokens_path).values()
        for token in [entity_type.start, entity_type.end]
        if token
    ]

    if existing:
        data = {
            "dataset-id": "dataset-id",
            "image": {
                "iiif_url": f"{FIXTURES}/extraction/images/text_line/test-page_1-line_1.jpg",
                "polygon": [
                    [37, 191],
                    [37, 339],
                    [767, 339],
                    [767, 191],
                    [37, 191],
                ],
            },
            "text": "%",
        }

        (output / "split.json").write_text(
            json.dumps({"train": {"train-page_1-line_5": data}})
        )

        split_content["train"]["train-page_1-line_5"] = data

    extractor = ArkindexExtractor(
        dataset_ids=["dataset_id"],
        element_type=["text_line"],
        output=output,
        # Keep the whole text
        entity_separators=None,
        transcription_worker_versions=[transcription_entities_worker_version],
        entity_worker_versions=[transcription_entities_worker_version]
        if load_entities
        else [],
        tokens=tokens_path if load_entities else None,
        keep_spaces=keep_spaces,
    )

    extractor.run()

    assert sorted(filter(methodcaller("is_file"), output.rglob("*"))) == [
        output / "split.json"
    ]

    split_content, _ = change_split_content(
        load_entities,
        transcription_entities_worker_version,
        keep_spaces,
        split_content,
        tokens,
    )

    assert json.loads((output / "split.json").read_text()) == split_content


@pytest.mark.parametrize("allow_empty", (True, False))
def test_empty_transcription(allow_empty, mock_database, tmp_path):
    output = tmp_path / "extraction"
    extractor = ArkindexExtractor(
        element_type=["text_line"],
        entity_separators=None,
        allow_empty=allow_empty,
        output=output,
    )
    element_no_transcription = Element(id="unknown")
    if allow_empty:
        assert extractor.extract_transcription(element_no_transcription) == ""
    else:
        with pytest.raises(NoTranscriptionError):
            extractor.extract_transcription(element_no_transcription)


@pytest.mark.parametrize("tokens", (None, EXTRACTION_DATA_PATH / "tokens.yml"))
def test_extract_transcription_no_translation(mock_database, tokens, tmp_path):
    output = tmp_path / "extraction"
    extractor = ArkindexExtractor(
        element_type=["text_line"],
        entity_separators=None,
        tokens=tokens,
        output=output,
    )

    element = Element.get_by_id("test-page_1-line_1")
    # Deleting one of the two transcriptions from the element
    Transcription.get(
        Transcription.element == element,
        Transcription.worker_run == "worker_run_id",
    ).delete_instance(recursive=True)

    # Deleting all entities on the element remaining transcription while leaving the transcription intact
    if tokens:
        TranscriptionEntity.delete().where(
            TranscriptionEntity.transcription
            == Transcription.select().where(Transcription.element == element).get()
        ).execute()

    # Early return with only the element transcription text instead of a translation
    assert extractor.extract_transcription(element) == "Leunaut  Claude  49"


@pytest.mark.parametrize(
    "nestation, xml_output, separators",
    (
        # Non-nested
        (
            "non-nested",
            "<root>The <adj>great</adj> king <name>Charles</name> III has eaten \nwith <person>us</person>.</root>",
            None,
        ),
        # Non-nested no text between entities
        (
            "non-nested",
            "<root><adj>great</adj> <name>Charles</name>\n<person>us</person></root>",
            ["\n", " "],
        ),
        # Nested
        (
            "nested",
            "<root>The great king <fullname><name>Charles</name> III</fullname> has eaten \nwith <person>us</person>.</root>",
            None,
        ),
        # Nested no text between entities
        (
            "nested",
            "<root><fullname><name>Charles</name> III</fullname>\n<person>us</person></root>",
            ["\n", " "],
        ),
        # Special characters in entities
        (
            "special-chars",
            "<root>The <Arkindex_s_entity>great</Arkindex_s_entity> king <_Name_1_>Charles</_Name_1_> III has eaten \nwith <Person_>us</Person_>.</root>",
            None,
        ),
    ),
)
def test_entities_to_xml(mock_database, nestation, xml_output, separators):
    transcription = Transcription.get_by_id("tr-with-entities")
    assert (
        entities_to_xml(
            text=transcription.text,
            predictions=get_transcription_entities(
                transcription_id="tr-with-entities",
                entity_worker_versions=[f"worker-version-{nestation}-id"],
                entity_worker_runs=[f"worker-run-{nestation}-id"],
                supported_types=[
                    "name",
                    "fullname",
                    "person",
                    "adj",
                    "Arkindex's entity",
                    '"Name" (1)',
                    "Person /!\\",
                ],
            ),
            entity_separators=separators,
        )
        == xml_output
    )


@pytest.mark.parametrize(
    "supported_entities, xml_output, separators",
    (
        # <adj> missing, no text between entities
        (
            ["name", "person"],
            "<root><name>Charles</name>\n<person>us</person></root>",
            ["\n", " "],
        ),
        # <adj> missing, text between entities
        (
            ["name", "person"],
            "<root>The great king <name>Charles</name> III has eaten \nwith <person>us</person>.</root>",
            None,
        ),
    ),
)
def test_entities_to_xml_partial_entities(
    mock_database, supported_entities, xml_output, separators
):
    transcription = Transcription.get_by_id("tr-with-entities")
    assert (
        entities_to_xml(
            text=transcription.text,
            predictions=get_transcription_entities(
                transcription_id="tr-with-entities",
                entity_worker_versions=["worker-version-non-nested-id"],
                entity_worker_runs=["worker-run-non-nested-id"],
                supported_types=supported_entities,
            ),
            entity_separators=separators,
        )
        == xml_output
    )


@pytest.mark.parametrize(
    "transcription",
    (
        "Something\n",
        "Something\r",
        "Something\t",
        'Something"',
        "Something'",
        "Something<",
        "Something>",
        "Something&",
    ),
)
def test_entities_to_xml_no_encode(transcription):
    assert (
        entities_to_xml(
            text=transcription,
            # Empty queryset
            predictions=TranscriptionEntity.select().where(TranscriptionEntity.id == 0),
            entity_separators=None,
        )
        == f"<root>{transcription}</root>"
    )



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import random

from torch.nn import (
    Conv2d,
    Dropout,
    Dropout2d,
    InstanceNorm2d,
    Module,
    ModuleList,
    ReLU,
)
from torch.nn.functional import pad


class DepthSepConv2D(Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        activation=None,
        padding=True,
        stride=(1, 1),
        dilation=(1, 1),
    ):
        super(DepthSepConv2D, self).__init__()

        self.padding = None

        if padding:
            if padding is True:
                padding = [int((k - 1) / 2) for k in kernel_size]
                if kernel_size[0] % 2 == 0 or kernel_size[1] % 2 == 0:
                    padding_h = kernel_size[1] - 1
                    padding_w = kernel_size[0] - 1
                    self.padding = [
                        padding_h // 2,
                        padding_h - padding_h // 2,
                        padding_w // 2,
                        padding_w - padding_w // 2,
                    ]
                    padding = (0, 0)

        else:
            padding = (0, 0)
        self.depth_conv = Conv2d(
            in_channels=in_channels,
            out_channels=in_channels,
            kernel_size=kernel_size,
            dilation=dilation,
            stride=stride,
            padding=padding,
            groups=in_channels,
        )
        self.point_conv = Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            dilation=dilation,
            kernel_size=(1, 1),
        )
        self.activation = activation

    def forward(self, x):
        x = self.depth_conv(x)
        if self.padding:
            x = pad(x, self.padding)
        if self.activation:
            x = self.activation(x)
        x = self.point_conv(x)
        return x


class MixDropout(Module):
    def __init__(self, dropout_proba=0.4, dropout2d_proba=0.2):
        super(MixDropout, self).__init__()

        self.dropout = Dropout(dropout_proba)
        self.dropout2d = Dropout2d(dropout2d_proba)

    def forward(self, x):
        if random.random() < 0.5:
            return self.dropout(x)
        return self.dropout2d(x)


class FCN_Encoder(Module):
    def __init__(self, params):
        super(FCN_Encoder, self).__init__()

        self.dropout = params["dropout"]

        self.init_blocks = ModuleList(
            [
                ConvBlock(3, 16, stride=(1, 1), dropout=self.dropout),
                ConvBlock(16, 32, stride=(2, 2), dropout=self.dropout),
                ConvBlock(32, 64, stride=(2, 2), dropout=self.dropout),
                ConvBlock(64, 128, stride=(2, 2), dropout=self.dropout),
                ConvBlock(128, 128, stride=(2, 1), dropout=self.dropout),
                ConvBlock(128, 128, stride=(2, 1), dropout=self.dropout),
            ]
        )
        self.blocks = ModuleList(
            [
                DSCBlock(128, 128, stride=(1, 1), dropout=self.dropout),
                DSCBlock(128, 128, stride=(1, 1), dropout=self.dropout),
                DSCBlock(128, 128, stride=(1, 1), dropout=self.dropout),
                DSCBlock(128, 256, stride=(1, 1), dropout=self.dropout),
            ]
        )

    def forward(self, x):
        for b in self.init_blocks:
            x = b(x)
        for b in self.blocks:
            xt = b(x)
            x = x + xt if x.size() == xt.size() else xt
        return x


class ConvBlock(Module):
    def __init__(self, in_, out_, stride=(1, 1), k=3, activation=ReLU, dropout=0.4):
        super(ConvBlock, self).__init__()

        self.activation = activation()
        self.conv1 = Conv2d(
            in_channels=in_, out_channels=out_, kernel_size=k, padding=k // 2
        )
        self.conv2 = Conv2d(
            in_channels=out_, out_channels=out_, kernel_size=k, padding=k // 2
        )
        self.conv3 = Conv2d(
            out_, out_, kernel_size=(3, 3), padding=(1, 1), stride=stride
        )
        self.norm_layer = InstanceNorm2d(
            out_, eps=0.001, momentum=0.99, track_running_stats=False
        )
        self.dropout = MixDropout(dropout_proba=dropout, dropout2d_proba=dropout / 2)

    def forward(self, x):
        pos = random.randint(1, 3)
        x = self.conv1(x)
        x = self.activation(x)

        if pos == 1:
            x = self.dropout(x)

        x = self.conv2(x)
        x = self.activation(x)

        if pos == 2:
            x = self.dropout(x)

        x = self.norm_layer(x)
        x = self.conv3(x)
        x = self.activation(x)

        if pos == 3:
            x = self.dropout(x)
        return x


class DSCBlock(Module):
    def __init__(self, in_, out_, stride=(2, 1), activation=ReLU, dropout=0.4):
        super(DSCBlock, self).__init__()

        self.activation = activation()
        self.conv1 = DepthSepConv2D(in_, out_, kernel_size=(3, 3))
        self.conv2 = DepthSepConv2D(out_, out_, kernel_size=(3, 3))
        self.conv3 = DepthSepConv2D(
            out_, out_, kernel_size=(3, 3), padding=(1, 1), stride=stride
        )
        self.norm_layer = InstanceNorm2d(
            out_, eps=0.001, momentum=0.99, track_running_stats=False
        )
        self.dropout = MixDropout(dropout_proba=dropout, dropout2d_proba=dropout / 2)

    def forward(self, x):
        pos = random.randint(1, 3)
        x = self.conv1(x)
        x = self.activation(x)

        if pos == 1:
            x = self.dropout(x)

        x = self.conv2(x)
        x = self.activation(x)

        if pos == 2:
            x = self.dropout(x)

        x = self.norm_layer(x)
        x = self.conv3(x)

        if pos == 3:
            x = self.dropout(x)
        return x



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
"""
Analyze dataset and display statistics in markdown format.
"""

from pathlib import Path

from dan.datasets.analyze.statistics import run
from dan.utils import read_yaml


def add_analyze_parser(subcommands) -> None:
    parser = subcommands.add_parser(
        "analyze",
        description=__doc__,
        help=__doc__,
    )
    parser.add_argument(
        "--labels",
        type=Path,
        help="Path to the formatted labels in JSON format.",
        required=True,
    )
    parser.add_argument(
        "--tokens",
        type=read_yaml,
        help="Path to the tokens YAML file.",
        required=False,
    )
    parser.add_argument(
        "--output-file",
        dest="output",
        type=Path,
        help="The statistics will be saved to this file in Markdown format.",
        required=True,
    )

    parser.set_defaults(func=run)



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import json
import logging
import random
from copy import deepcopy

import numpy as np
import torch
import torch.multiprocessing as mp

from dan.ocr import wandb
from dan.ocr.manager.training import Manager
from dan.ocr.mlflow import MLFLOW_AVAILABLE
from dan.ocr.utils import update_config
from dan.utils import MLflowNotInstalled

if MLFLOW_AVAILABLE:
    import mlflow

    from dan.ocr.mlflow import make_mlflow_request, start_mlflow_run


logger = logging.getLogger(__name__)


def train(rank, params, mlflow_logging=False):
    torch.manual_seed(0)
    torch.cuda.manual_seed(0)
    np.random.seed(0)
    random.seed(0)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True

    params["training"]["device"]["ddp_rank"] = rank
    model = Manager(params)
    model.load_model()

    if params["dataset"]["tokens"] is not None:
        if "ner" not in params["training"]["metrics"]["train"]:
            params["training"]["metrics"]["train"].append("ner")
        if "ner" not in params["training"]["metrics"]["eval"]:
            params["training"]["metrics"]["eval"].append("ner")

    if mlflow_logging:
        logger.info("MLflow logging enabled")

    model.train(mlflow_logging=mlflow_logging)


def serialize_config(config):
    """
    Make every field of the configuration JSON-Serializable and remove sensitive information.

    - Classes are transformed using their name attribute
    - Functions are casted to strings
    """
    # Create a copy of the original config without erase it
    serialized_config = deepcopy(config)

    # Remove credentials to the config
    serialized_config["mlflow"]["s3_endpoint_url"] = ""
    serialized_config["mlflow"]["tracking_uri"] = ""
    serialized_config["mlflow"]["aws_access_key_id"] = ""
    serialized_config["mlflow"]["aws_secret_access_key"] = ""

    # Get the name of the class
    serialized_config["model"]["models"]["encoder"] = serialized_config["model"][
        "models"
    ]["encoder"].__name__
    serialized_config["model"]["models"]["decoder"] = serialized_config["model"][
        "models"
    ]["decoder"].__name__
    serialized_config["training"]["optimizers"]["all"]["class"] = serialized_config[
        "training"
    ]["optimizers"]["all"]["class"].__name__

    # Cast the functions to str
    serialized_config["dataset"]["config"]["augmentation"] = str(
        serialized_config["dataset"]["config"]["augmentation"]
    )
    serialized_config["training"]["nb_gpu"] = str(
        serialized_config["training"]["nb_gpu"]
    )

    return serialized_config


def start_training(config, mlflow_logging: bool) -> None:
    if (
        config["training"]["device"]["use_ddp"]
        and config["training"]["device"]["force"] in [None, "cuda"]
        and torch.cuda.is_available()
    ):
        mp.spawn(
            train,
            args=(config, mlflow_logging),
            nprocs=config["training"]["device"]["nb_gpu"],
        )
    else:
        train(0, config, mlflow_logging)


def run(config: dict):
    """
    Main program, training a new model, using a valid configuration
    """
    names = list(config["dataset"]["datasets"].keys())
    # We should only have one dataset
    assert len(names) == 1, f"Found {len(names)} datasets but only one is expected"

    dataset_name = names.pop()
    update_config(config)

    # Start "Weights & Biases" as soon as possible
    wandb.init(config)

    if config.get("mlflow") and not MLFLOW_AVAILABLE:
        logger.error(
            "Cannot log to MLflow. Please install the `mlflow` extra requirements."
        )
        raise MLflowNotInstalled()

    if not config.get("mlflow"):
        start_training(config, mlflow_logging=False)
    else:
        labels_path = config["dataset"]["datasets"][dataset_name] / "labels.json"
        with start_mlflow_run(config["mlflow"]) as (run, created):
            if created:
                logger.info(f"Started MLflow run with ID ({run.info.run_id})")
            else:
                logger.info(f"Resumed MLflow run with ID ({run.info.run_id})")

            make_mlflow_request(
                mlflow_method=mlflow.set_tags, tags={"Dataset": dataset_name}
            )
            # Get the labels json file
            labels_artifact = json.loads(labels_path.read_text())

            # Log MLflow artifacts
            for artifact, filename in [
                (serialize_config(config), "config.json"),
                (labels_artifact, "labels.json"),
            ]:
                make_mlflow_request(
                    mlflow_method=mlflow.log_dict,
                    dictionary=artifact,
                    artifact_file=filename,
                )
            start_training(config, mlflow_logging=True)



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

import json
import logging
import random
from collections import defaultdict
from pathlib import Path
from typing import Dict, List
from uuid import UUID

from tqdm import tqdm

from arkindex_export import Dataset, DatasetElement, Element, open_database
from dan.datasets.extract.db import (
    get_dataset_elements,
    get_elements,
    get_transcription_entities,
    get_transcriptions,
)
from dan.datasets.extract.exceptions import (
    NoTranscriptionError,
    ProcessingError,
)
from dan.datasets.extract.utils import (
    entities_to_xml,
    get_translation_map,
    normalize_linebreaks,
    normalize_spaces,
)
from dan.utils import parse_tokens

TRAIN_NAME = "train"
VAL_NAME = "val"
TEST_NAME = "test"
SPLIT_NAMES = [TRAIN_NAME, VAL_NAME, TEST_NAME]

logger = logging.getLogger(__name__)


class ArkindexExtractor:
    """
    Extract data from Arkindex
    """

    def __init__(
        self,
        output: Path,
        dataset_ids: List[UUID] | None = None,
        element_type: List[str] = [],
        entity_separators: List[str] = ["\n", " "],
        tokens: Path | None = None,
        transcription_worker_versions: List[str | bool] = [],
        entity_worker_versions: List[str | bool] = [],
        transcription_worker_runs: List[str | bool] = [],
        entity_worker_runs: List[str | bool] = [],
        keep_spaces: bool = False,
        allow_empty: bool = False,
    ) -> None:
        self.dataset_ids = dataset_ids
        self.element_type = element_type
        self.output = output
        self.entity_separators = entity_separators
        self.tokens = parse_tokens(tokens) if tokens else {}
        self.transcription_worker_versions = transcription_worker_versions
        self.entity_worker_versions = entity_worker_versions
        self.transcription_worker_runs = transcription_worker_runs
        self.entity_worker_runs = entity_worker_runs
        self.allow_empty = allow_empty
        self.keep_spaces = keep_spaces

        data_path = self.output / "split.json"
        # New keys can appear between several extractions
        # We must explicitly define that this dict expects a dict as its value
        self.data = defaultdict(dict)
        if data_path.exists():
            self.data.update(json.loads(data_path.read_text()))

        # NER extraction
        self.translation_map: Dict[str, str] | None = get_translation_map(self.tokens)

    def translate(self, text: str):
        """
        Use translation map to replace XML tags to actual tokens
        """
        for pattern, repl in self.translation_map.items():
            text = text.replace(pattern, repl)
        return text

    def extract_transcription(self, element: Element):
        """
        Extract the element's transcription.
        If the entities are needed, they are added to the transcription using tokens.
        """
        transcriptions = get_transcriptions(
            element.id,
            self.transcription_worker_versions,
            self.transcription_worker_runs,
        )
        if len(transcriptions) == 0:
            if self.allow_empty:
                return ""
            raise NoTranscriptionError(element.id)

        transcription = random.choice(transcriptions)
        stripped_text = transcription.text.strip()

        if not self.tokens:
            return stripped_text

        entities = get_transcription_entities(
            transcription.id,
            self.entity_worker_versions,
            self.entity_worker_runs,
            supported_types=list(self.tokens),
        )

        if not entities.count():
            return stripped_text

        return self.translate(
            entities_to_xml(
                transcription.text, entities, entity_separators=self.entity_separators
            )
        )

    def format_text(self, text: str):
        if not self.keep_spaces:
            text = normalize_spaces(text)
            text = normalize_linebreaks(text)

        return text.strip()

    def process_element(self, dataset_parent: DatasetElement, element: Element):
        """
        Extract an element's data and save it to disk.
        The output path is directly related to the split of the element.
        """
        text = self.extract_transcription(element)
        text = self.format_text(text)

        self.data[dataset_parent.set_name][element.id] = {
            "dataset_id": dataset_parent.dataset_id,
            "text": text,
            "image": {
                "iiif_url": element.image.url,
                "polygon": json.loads(element.polygon),
            },
        }

    def process_parent(self, pbar, dataset_parent: DatasetElement):
        """
        Extract data from a parent element.
        """
        parent = dataset_parent.element
        base_description = f"Extracting data from {parent.type} ({parent.id}) for split ({dataset_parent.set_name})"
        pbar.set_description(desc=base_description)
        if self.element_type == [parent.type]:
            try:
                self.process_element(dataset_parent, parent)
            except ProcessingError as e:
                logger.warning(f"Skipping {parent.id}: {str(e)}")
        # Extract children elements
        else:
            children = get_elements(
                parent.id,
                self.element_type,
            )

            nb_children = children.count()
            for idx, element in enumerate(children, start=1):
                # Update description to update the children processing progress
                pbar.set_description(desc=base_description + f" ({idx}/{nb_children})")
                try:
                    self.process_element(dataset_parent, element)
                except ProcessingError as e:
                    logger.warning(f"Skipping {element.id}: {str(e)}")

    def export(self):
        (self.output / "split.json").write_text(
            json.dumps(
                self.data,
                sort_keys=True,
                indent=4,
            )
        )

    def run(self):
        # Retrieve the Dataset and its splits from the cache
        for dataset_id in self.dataset_ids:
            dataset = Dataset.get_by_id(dataset_id)
            splits = dataset.sets.split(",")
            if not set(splits).issubset(set(SPLIT_NAMES)):
                logger.warning(
                    f'Dataset {dataset.name} ({dataset.id}) does not have "{TRAIN_NAME}", "{VAL_NAME}" and "{TEST_NAME}" steps'
                )
                continue

            # Extract the train set first to correctly build the `self.charset` variable
            splits.remove(TRAIN_NAME)
            splits.insert(0, TRAIN_NAME)

            # Iterate over the subsets to find the page images and labels.
            for split in splits:
                with tqdm(
                    get_dataset_elements(dataset, split),
                    desc=f"Extracting data from ({dataset_id}) for split ({split})",
                ) as pbar:
                    # Iterate over the pages to create splits at page level.
                    for parent in pbar:
                        self.process_parent(
                            pbar=pbar,
                            dataset_parent=parent,
                        )
                        # Progress bar updates
                        pbar.update()
                        pbar.refresh()

        if not self.data:
            raise Exception(
                "No data was extracted using the provided export database and parameters."
            )

        self.export()


def run(
    database: Path,
    dataset_ids: List[UUID],
    element_type: List[str],
    output: Path,
    entity_separators: List[str],
    tokens: Path,
    transcription_worker_versions: List[str | bool],
    entity_worker_versions: List[str | bool],
    transcription_worker_runs: List[str | bool],
    entity_worker_runs: List[str | bool],
    keep_spaces: bool,
    allow_empty: bool,
):
    assert database.exists(), f"No file found @ {database}"
    open_database(path=database)

    # Create directories
    output.mkdir(parents=True, exist_ok=True)

    ArkindexExtractor(
        dataset_ids=dataset_ids,
        element_type=element_type,
        output=output,
        entity_separators=entity_separators,
        tokens=tokens,
        transcription_worker_versions=transcription_worker_versions,
        entity_worker_versions=entity_worker_versions,
        transcription_worker_runs=transcription_worker_runs,
        entity_worker_runs=entity_worker_runs,
        keep_spaces=keep_spaces,
        allow_empty=allow_empty,
    ).run()



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

import json
import shutil
from pathlib import Path

import numpy as np
import pytest
import yaml

from dan.ocr.predict.attention import Level
from dan.ocr.predict.inference import DAN
from dan.ocr.predict.inference import run as run_prediction
from dan.utils import (
    parse_charset_pattern,
    parse_tokens,
    read_yaml,
)
from tests import FIXTURES

PREDICTION_DATA_PATH = FIXTURES / "prediction"

FONT_PATH = FIXTURES.parent.parent / "fonts/LinuxLibertine.ttf"
MAXIMUM_FONT_SIZE = 32


@pytest.mark.parametrize(
    "image_name, expected_prediction",
    (
        (
            "0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84.png",
            {"text": ["ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241"]},
        ),
        (
            "0dfe8bcd-ed0b-453e-bf19-cc697012296e.png",
            {"text": ["ⓈTemplié ⒻMarcelle Ⓑ93 ⓁS Ⓚch ⓄE dactylo Ⓟ18376"]},
        ),
        (
            "2c242f5c-e979-43c4-b6f2-a6d4815b651d.png",
            {"text": ["Ⓢd ⒻCharles Ⓑ11 ⓁP ⒸC ⓀF Ⓞd Ⓟ14 31"]},
        ),
        (
            "ffdec445-7f14-4f5f-be44-68d0844d0df1.png",
            {"text": ["ⓈNaudin ⒻMarie Ⓑ53 ⓁS Ⓒv ⓀBelle mère"]},
        ),
    ),
)
@pytest.mark.parametrize("normalize", (True, False))
def test_predict(image_name, expected_prediction, normalize, tmp_path):
    # Update mean/std in parameters.yml
    model_path = tmp_path / "models"
    model_path.mkdir(exist_ok=True)

    shutil.copyfile(
        PREDICTION_DATA_PATH / "model.pt",
        model_path / "model.pt",
    )
    shutil.copyfile(
        PREDICTION_DATA_PATH / "charset.pkl",
        model_path / "charset.pkl",
    )

    params = read_yaml(PREDICTION_DATA_PATH / "parameters.yml")
    if not normalize:
        del params["parameters"]["mean"]
        del params["parameters"]["std"]
    yaml.dump(params, (model_path / "parameters.yml").open("w"))

    device = "cpu"

    dan_model = DAN(device)
    dan_model.load(path=model_path, mode="eval")

    image_path = PREDICTION_DATA_PATH / "images" / image_name
    _, image = dan_model.preprocess(str(image_path))

    input_tensor = image.unsqueeze(0)
    input_tensor = input_tensor.to(device)
    input_sizes = [image.shape[1:]]

    prediction = dan_model.predict(
        input_tensor,
        input_sizes,
        char_separators=parse_charset_pattern(dan_model.charset),
    )

    assert prediction == expected_prediction


@pytest.mark.parametrize(
    "image_name, confidence_score, temperature, predict_objects, expected_prediction",
    (
        (
            "0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84",
            [],  # Confidence score
            1.0,  # Temperature
            False,  # Predict objects
            {
                "text": "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                "language_model": {},
                "confidences": {},
            },
        ),
        (
            "0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84",
            [Level.Word],  # Confidence score
            1.0,  # Temperature
            True,  # Predict objects
            {
                "text": "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                "language_model": {},
                "confidences": {
                    "total": 1.0,
                    "word": [
                        {"text": "ⓈBellisson", "confidence": 1.0},
                        {"text": "ⒻGeorges", "confidence": 1.0},
                        {"text": "Ⓑ91", "confidence": 1.0},
                        {"text": "ⓁP", "confidence": 1.0},
                        {"text": "ⒸM", "confidence": 1.0},
                        {"text": "ⓀCh", "confidence": 1.0},
                        {"text": "ⓄPlombier", "confidence": 1.0},
                        {"text": "ⓅPatron?12241", "confidence": 1.0},
                    ],
                },
                "objects": [
                    {
                        "confidence": 0.42,
                        "polygon": [[0, 0], [144, 0], [144, 66], [0, 66]],
                        "text": "ⓈBellisson",
                        "text_confidence": 1.0,
                    },
                    {
                        "confidence": 0.52,
                        "polygon": [[184, 0], [269, 0], [269, 66], [184, 66]],
                        "text": "ⒻGeorges",
                        "text_confidence": 1.0,
                    },
                    {
                        "confidence": 0.21,
                        "polygon": [[294, 0], [371, 0], [371, 66], [294, 66]],
                        "text": "Ⓑ91",
                        "text_confidence": 1.0,
                    },
                    {
                        "confidence": 0.23,
                        "polygon": [[367, 0], [427, 0], [427, 66], [367, 66]],
                        "text": "ⓁP",
                        "text_confidence": 1.0,
                    },
                    {
                        "confidence": 0.18,
                        "polygon": [[535, 0], [619, 0], [619, 66], [535, 66]],
                        "text": "ⒸM",
                        "text_confidence": 1.0,
                    },
                    {
                        "confidence": 0.23,
                        "polygon": [[589, 0], [674, 0], [674, 66], [589, 66]],
                        "text": "ⓀCh",
                        "text_confidence": 1.0,
                    },
                    {
                        "confidence": 0.31,
                        "polygon": [[685, 0], [806, 0], [806, 66], [685, 66]],
                        "text": "ⓄPlombier",
                        "text_confidence": 1.0,
                    },
                    {
                        "confidence": 0.91,
                        "polygon": [[820, 0], [938, 0], [938, 66], [820, 66]],
                        "text": "ⓅPatron?12241",
                        "text_confidence": 1.0,
                    },
                ],
                "attention_gif": "0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84_word.gif",
            },
        ),
        (
            "0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84",
            [Level.NER, Level.Word],  # Confidence score
            3.5,  # Temperature
            False,  # Predict objects
            {
                "text": "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                "language_model": {},
                "confidences": {
                    "total": 0.93,
                    "ner": [
                        {"text": "ⓈBellisson ", "confidence": 0.92},
                        {"text": "ⒻGeorges ", "confidence": 0.94},
                        {"text": "Ⓑ91 ", "confidence": 0.93},
                        {"text": "ⓁP ", "confidence": 0.92},
                        {"text": "ⒸM ", "confidence": 0.93},
                        {"text": "ⓀCh ", "confidence": 0.95},
                        {"text": "ⓄPlombier ", "confidence": 0.93},
                        {"text": "ⓅPatron?12241", "confidence": 0.93},
                    ],
                    "word": [
                        {"text": "ⓈBellisson", "confidence": 0.93},
                        {"text": "ⒻGeorges", "confidence": 0.94},
                        {"text": "Ⓑ91", "confidence": 0.92},
                        {"text": "ⓁP", "confidence": 0.94},
                        {"text": "ⒸM", "confidence": 0.93},
                        {"text": "ⓀCh", "confidence": 0.96},
                        {"text": "ⓄPlombier", "confidence": 0.94},
                        {"text": "ⓅPatron?12241", "confidence": 0.93},
                    ],
                },
            },
        ),
        (
            "0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84",
            [Level.Line],  # Confidence score
            1.0,  # Temperature
            False,  # Predict objects
            {
                "text": "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                "language_model": {},
                "confidences": {
                    "total": 1.0,
                    "line": [
                        {
                            "text": "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                            "confidence": 1.0,
                        }
                    ],
                },
            },
        ),
        (
            "0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84",
            [Level.NER, Level.Line],  # Confidence score
            3.5,  # Temperature
            False,  # Predict objects
            {
                "text": "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                "language_model": {},
                "confidences": {
                    "total": 0.93,
                    "ner": [
                        {"text": "ⓈBellisson ", "confidence": 0.92},
                        {"text": "ⒻGeorges ", "confidence": 0.94},
                        {"text": "Ⓑ91 ", "confidence": 0.93},
                        {"text": "ⓁP ", "confidence": 0.92},
                        {"text": "ⒸM ", "confidence": 0.93},
                        {"text": "ⓀCh ", "confidence": 0.95},
                        {"text": "ⓄPlombier ", "confidence": 0.93},
                        {"text": "ⓅPatron?12241", "confidence": 0.93},
                    ],
                    "line": [
                        {
                            "text": "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                            "confidence": 0.93,
                        }
                    ],
                },
            },
        ),
        (
            "0dfe8bcd-ed0b-453e-bf19-cc697012296e",
            [],  # Confidence score
            1.0,  # Temperature
            False,  # Predict objects
            {
                "text": "ⓈTemplié ⒻMarcelle Ⓑ93 ⓁS Ⓚch ⓄE dactylo Ⓟ18376",
                "language_model": {},
                "confidences": {},
            },
        ),
        (
            "0dfe8bcd-ed0b-453e-bf19-cc697012296e",
            [Level.NER, Level.Char, Level.Word, Level.Line],  # Confidence score
            1.0,  # Temperature
            False,  # Predict objects
            {
                "text": "ⓈTemplié ⒻMarcelle Ⓑ93 ⓁS Ⓚch ⓄE dactylo Ⓟ18376",
                "language_model": {},
                "confidences": {
                    "total": 1.0,
                    "ner": [
                        {"text": "ⓈTemplié ", "confidence": 0.98},
                        {"text": "ⒻMarcelle ", "confidence": 1.0},
                        {"text": "Ⓑ93 ", "confidence": 1.0},
                        {"text": "ⓁS ", "confidence": 1.0},
                        {"text": "Ⓚch ", "confidence": 1.0},
                        {"text": "ⓄE dactylo ", "confidence": 1.0},
                        {"text": "Ⓟ18376", "confidence": 1.0},
                    ],
                    "char": [
                        {"text": "Ⓢ", "confidence": 1.0},
                        {"text": "T", "confidence": 1.0},
                        {"text": "e", "confidence": 1.0},
                        {"text": "m", "confidence": 1.0},
                        {"text": "p", "confidence": 1.0},
                        {"text": "l", "confidence": 1.0},
                        {"text": "i", "confidence": 1.0},
                        {"text": "é", "confidence": 0.85},
                        {"text": " ", "confidence": 1.0},
                        {"text": "Ⓕ", "confidence": 1.0},
                        {"text": "M", "confidence": 1.0},
                        {"text": "a", "confidence": 1.0},
                        {"text": "r", "confidence": 1.0},
                        {"text": "c", "confidence": 1.0},
                        {"text": "e", "confidence": 1.0},
                        {"text": "l", "confidence": 1.0},
                        {"text": "l", "confidence": 1.0},
                        {"text": "e", "confidence": 1.0},
                        {"text": " ", "confidence": 1.0},
                        {"text": "Ⓑ", "confidence": 1.0},
                        {"text": "9", "confidence": 1.0},
                        {"text": "3", "confidence": 1.0},
                        {"text": " ", "confidence": 1.0},
                        {"text": "Ⓛ", "confidence": 1.0},
                        {"text": "S", "confidence": 1.0},
                        {"text": " ", "confidence": 1.0},
                        {"text": "Ⓚ", "confidence": 1.0},
                        {"text": "c", "confidence": 1.0},
                        {"text": "h", "confidence": 1.0},
                        {"text": " ", "confidence": 1.0},
                        {"text": "Ⓞ", "confidence": 1.0},
                        {"text": "E", "confidence": 1.0},
                        {"text": " ", "confidence": 1.0},
                        {"text": "d", "confidence": 1.0},
                        {"text": "a", "confidence": 1.0},
                        {"text": "c", "confidence": 1.0},
                        {"text": "t", "confidence": 1.0},
                        {"text": "y", "confidence": 1.0},
                        {"text": "l", "confidence": 1.0},
                        {"text": "o", "confidence": 1.0},
                        {"text": " ", "confidence": 1.0},
                        {"text": "Ⓟ", "confidence": 1.0},
                        {"text": "1", "confidence": 1.0},
                        {"text": "8", "confidence": 1.0},
                        {"text": "3", "confidence": 1.0},
                        {"text": "7", "confidence": 1.0},
                        {"text": "6", "confidence": 1.0},
                    ],
                    "word": [
                        {"text": "ⓈTemplié", "confidence": 0.98},
                        {"text": "ⒻMarcelle", "confidence": 1.0},
                        {"text": "Ⓑ93", "confidence": 1.0},
                        {"text": "ⓁS", "confidence": 1.0},
                        {"text": "Ⓚch", "confidence": 1.0},
                        {"text": "ⓄE", "confidence": 1.0},
                        {"text": "dactylo", "confidence": 1.0},
                        {"text": "Ⓟ18376", "confidence": 1.0},
                    ],
                    "line": [
                        {
                            "text": "ⓈTemplié ⒻMarcelle Ⓑ93 ⓁS Ⓚch ⓄE dactylo Ⓟ18376",
                            "confidence": 1.0,
                        }
                    ],
                },
            },
        ),
        (
            "2c242f5c-e979-43c4-b6f2-a6d4815b651d",
            [],  # Confidence score
            1.0,  # Temperature
            False,  # Predict objects
            {
                "text": "Ⓢd ⒻCharles Ⓑ11 ⓁP ⒸC ⓀF Ⓞd Ⓟ14 31",
                "language_model": {},
                "confidences": {},
            },
        ),
        (
            "ffdec445-7f14-4f5f-be44-68d0844d0df1",
            [],  # Confidence score
            1.0,  # Temperature
            True,  # Predict objects
            {
                "text": "ⓈNaudin ⒻMarie Ⓑ53 ⓁS Ⓒv ⓀBelle mère",
                "language_model": {},
                "confidences": {},
                "objects": [
                    {
                        "confidence": 0.96,
                        "polygon": [[546, 0], [715, 0], [715, 67], [546, 67]],
                        "text": "ⓈNaudin ⒻMarie Ⓑ53 ⓁS Ⓒv ⓀBelle mère",
                        "text_confidence": 1.0,
                    }
                ],
            },
        ),
    ),
)
def test_run_prediction(
    image_name,
    confidence_score,
    temperature,
    predict_objects,
    expected_prediction,
    tmp_path,
):
    if "attention_gif" in expected_prediction:
        expected_prediction["attention_gif"] = str(
            tmp_path / expected_prediction["attention_gif"]
        )

    # Make tmpdir and copy needed image inside
    image_dir = tmp_path / "images"
    image_dir.mkdir()
    shutil.copyfile(
        (PREDICTION_DATA_PATH / "images" / image_name).with_suffix(".png"),
        (image_dir / image_name).with_suffix(".png"),
    )

    run_prediction(
        image_dir=image_dir,
        font=FONT_PATH,
        maximum_font_size=MAXIMUM_FONT_SIZE,
        model=PREDICTION_DATA_PATH,
        output=tmp_path,
        confidence_score=bool(confidence_score),
        confidence_score_levels=confidence_score,
        attention_map=predict_objects and confidence_score,
        attention_map_level=[Level.Line, *confidence_score].pop(),
        attention_map_scale=0.5,
        alpha_factor=0.9,
        color_map="nipy_spectral",
        word_separators=[" ", "\n"],
        line_separators=["\n"],
        temperature=temperature,
        predict_objects=predict_objects,
        max_object_height=None,
        image_extension=".png",
        gpu_device=None,
        batch_size=1,
        tokens=parse_tokens(PREDICTION_DATA_PATH / "tokens.yml"),
        start_token=None,
        use_language_model=False,
        compile_model=False,
        dynamic_mode=False,
    )

    prediction = json.loads((tmp_path / image_name).with_suffix(".json").read_text())
    assert prediction == expected_prediction
    if "attention_gif" in expected_prediction:
        assert Path(expected_prediction["attention_gif"]).exists()


@pytest.mark.parametrize(
    "image_names, confidence_score, temperature, expected_predictions",
    (
        (
            ["0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84"],
            None,
            1.0,
            [
                {
                    "text": "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                    "language_model": {},
                    "confidences": {},
                }
            ],
        ),
        (
            ["0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84"],
            [Level.Word],
            1.0,
            [
                {
                    "text": "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                    "language_model": {},
                    "confidences": {
                        "total": 1.0,
                        "word": [
                            {"text": "ⓈBellisson", "confidence": 1.0},
                            {"text": "ⒻGeorges", "confidence": 1.0},
                            {"text": "Ⓑ91", "confidence": 1.0},
                            {"text": "ⓁP", "confidence": 1.0},
                            {"text": "ⒸM", "confidence": 1.0},
                            {"text": "ⓀCh", "confidence": 1.0},
                            {"text": "ⓄPlombier", "confidence": 1.0},
                            {"text": "ⓅPatron?12241", "confidence": 1.0},
                        ],
                    },
                }
            ],
        ),
        (
            [
                "0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84",
                "0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84",
            ],
            [Level.NER, Level.Word],
            1.0,
            [
                {
                    "text": "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                    "language_model": {},
                    "confidences": {
                        "total": 1.0,
                        "ner": [
                            {"text": "ⓈBellisson ", "confidence": 1.0},
                            {"text": "ⒻGeorges ", "confidence": 1.0},
                            {"text": "Ⓑ91 ", "confidence": 1.0},
                            {"text": "ⓁP ", "confidence": 1.0},
                            {"text": "ⒸM ", "confidence": 1.0},
                            {"text": "ⓀCh ", "confidence": 1.0},
                            {"text": "ⓄPlombier ", "confidence": 1.0},
                            {"text": "ⓅPatron?12241", "confidence": 1.0},
                        ],
                        "word": [
                            {"text": "ⓈBellisson", "confidence": 1.0},
                            {"text": "ⒻGeorges", "confidence": 1.0},
                            {"text": "Ⓑ91", "confidence": 1.0},
                            {"text": "ⓁP", "confidence": 1.0},
                            {"text": "ⒸM", "confidence": 1.0},
                            {"text": "ⓀCh", "confidence": 1.0},
                            {"text": "ⓄPlombier", "confidence": 1.0},
                            {"text": "ⓅPatron?12241", "confidence": 1.0},
                        ],
                    },
                },
                {
                    "text": "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                    "language_model": {},
                    "confidences": {
                        "total": 1.0,
                        "ner": [
                            {"text": "ⓈBellisson ", "confidence": 1.0},
                            {"text": "ⒻGeorges ", "confidence": 1.0},
                            {"text": "Ⓑ91 ", "confidence": 1.0},
                            {"text": "ⓁP ", "confidence": 1.0},
                            {"text": "ⒸM ", "confidence": 1.0},
                            {"text": "ⓀCh ", "confidence": 1.0},
                            {"text": "ⓄPlombier ", "confidence": 1.0},
                            {"text": "ⓅPatron?12241", "confidence": 1.0},
                        ],
                        "word": [
                            {"text": "ⓈBellisson", "confidence": 1.0},
                            {"text": "ⒻGeorges", "confidence": 1.0},
                            {"text": "Ⓑ91", "confidence": 1.0},
                            {"text": "ⓁP", "confidence": 1.0},
                            {"text": "ⒸM", "confidence": 1.0},
                            {"text": "ⓀCh", "confidence": 1.0},
                            {"text": "ⓄPlombier", "confidence": 1.0},
                            {"text": "ⓅPatron?12241", "confidence": 1.0},
                        ],
                    },
                },
            ],
        ),
        (
            ["0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84"],
            [Level.Word],
            1.0,
            [
                {
                    "text": "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                    "language_model": {},
                    "confidences": {
                        "total": 1.0,
                        "word": [
                            {"text": "ⓈBellisson", "confidence": 1.0},
                            {"text": "ⒻGeorges", "confidence": 1.0},
                            {"text": "Ⓑ91", "confidence": 1.0},
                            {"text": "ⓁP", "confidence": 1.0},
                            {"text": "ⒸM", "confidence": 1.0},
                            {"text": "ⓀCh", "confidence": 1.0},
                            {"text": "ⓄPlombier", "confidence": 1.0},
                            {"text": "ⓅPatron?12241", "confidence": 1.0},
                        ],
                    },
                }
            ],
        ),
        (
            [
                "2c242f5c-e979-43c4-b6f2-a6d4815b651d",
                "ffdec445-7f14-4f5f-be44-68d0844d0df1",
            ],
            False,
            1.0,
            [
                {
                    "text": "Ⓢd ⒻCharles Ⓑ11 ⓁP ⒸC ⓀF Ⓞd Ⓟ14 31",
                    "language_model": {},
                    "confidences": {},
                },
                {
                    "text": "ⓈNaudin ⒻMarie Ⓑ53 ⓁS Ⓒv ⓀBelle mère",
                    "language_model": {},
                    "confidences": {},
                },
            ],
        ),
    ),
)
@pytest.mark.parametrize("batch_size", [1, 2])
def test_run_prediction_batch(
    image_names,
    confidence_score,
    temperature,
    expected_predictions,
    batch_size,
    tmp_path,
):
    # Make tmpdir and copy needed images inside
    image_dir = tmp_path / "images"
    image_dir.mkdir()
    for image_name in image_names:
        shutil.copyfile(
            (PREDICTION_DATA_PATH / "images" / image_name).with_suffix(".png"),
            (image_dir / image_name).with_suffix(".png"),
        )

    run_prediction(
        image_dir=image_dir,
        font=FONT_PATH,
        maximum_font_size=MAXIMUM_FONT_SIZE,
        model=PREDICTION_DATA_PATH,
        output=tmp_path,
        confidence_score=True if confidence_score else False,
        confidence_score_levels=confidence_score if confidence_score else [],
        attention_map=False,
        attention_map_level=None,
        attention_map_scale=0.5,
        alpha_factor=0.9,
        color_map="nipy_spectral",
        word_separators=[" ", "\n"],
        line_separators=["\n"],
        temperature=temperature,
        predict_objects=False,
        max_object_height=None,
        image_extension=".png",
        gpu_device=None,
        batch_size=batch_size,
        tokens=parse_tokens(PREDICTION_DATA_PATH / "tokens.yml"),
        start_token=None,
        use_language_model=False,
        compile_model=False,
        dynamic_mode=False,
    )

    for image_name, expected_prediction in zip(image_names, expected_predictions):
        prediction = json.loads(
            (tmp_path / image_name).with_suffix(".json").read_text()
        )
        assert prediction == expected_prediction


@pytest.mark.parametrize(
    "image_names, language_model_weight, expected_predictions",
    (
        (
            [
                "0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84",
                "0dfe8bcd-ed0b-453e-bf19-cc697012296e",
                "2c242f5c-e979-43c4-b6f2-a6d4815b651d",
                "ffdec445-7f14-4f5f-be44-68d0844d0df1",
            ],
            1.0,
            [
                {
                    "text": "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                    "language_model": {
                        "text": "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                        "confidence": 0.92,
                    },
                },
                {
                    "text": "ⓈTemplié ⒻMarcelle Ⓑ93 ⓁS Ⓚch ⓄE dactylo Ⓟ18376",
                    "language_model": {
                        "text": "ⓈTemplié ⒻMarcelle Ⓑ93 ⓁS Ⓚch ⓄE dactylo Ⓟ18376",
                        "confidence": 0.88,
                    },
                },
                {
                    "text": "Ⓢd ⒻCharles Ⓑ11 ⓁP ⒸC ⓀF Ⓞd Ⓟ14 31",
                    "language_model": {
                        "text": "Ⓢd ⒻCharles Ⓑ11 ⓁP ⒸC ⓀF Ⓞd Ⓟ14 31",
                        "confidence": 0.86,
                    },
                },
                {
                    "text": "ⓈNaudin ⒻMarie Ⓑ53 ⓁS Ⓒv ⓀBelle mère",
                    "language_model": {
                        "text": "ⓈNaudin ⒻMarie Ⓑ53 ⓁS Ⓒv ⓀBelle mère",
                        "confidence": 0.89,
                    },
                },
            ],
        ),
        (
            [
                "0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84",
                "0dfe8bcd-ed0b-453e-bf19-cc697012296e",
                "2c242f5c-e979-43c4-b6f2-a6d4815b651d",
                "ffdec445-7f14-4f5f-be44-68d0844d0df1",
            ],
            2.0,
            [
                {
                    "text": "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                    "language_model": {
                        "text": "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                        "confidence": 0.90,
                    },
                },
                {
                    "text": "ⓈTemplié ⒻMarcelle Ⓑ93 ⓁS Ⓚch ⓄE dactylo Ⓟ18376",
                    "language_model": {
                        "text": "ⓈTemplié ⒻMarcelle Ⓑ93 ⓁS Ⓚch ⓄE dactylo Ⓟ18376",
                        "confidence": 0.84,
                    },
                },
                {
                    "text": "Ⓢd ⒻCharles Ⓑ11 ⓁP ⒸC ⓀF Ⓞd Ⓟ14 31",
                    "language_model": {
                        "text": "Ⓢd ⒻCharles Ⓑ11 ⓁP ⒸC ⓀF Ⓞd Ⓟ14331",
                        "confidence": 0.83,
                    },
                },
                {
                    "text": "ⓈNaudin ⒻMarie Ⓑ53 ⓁS Ⓒv ⓀBelle mère",
                    "language_model": {
                        "text": "ⓈNaudin ⒻMarie Ⓑ53 ⓁS Ⓒv ⓀBelle mère",
                        "confidence": 0.86,
                    },
                },
            ],
        ),
        (
            [
                "0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84",
                "0dfe8bcd-ed0b-453e-bf19-cc697012296e",
                "2c242f5c-e979-43c4-b6f2-a6d4815b651d",
                "ffdec445-7f14-4f5f-be44-68d0844d0df1",
            ],
            0.0,
            [
                {"text": "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241"},
                {"text": "ⓈTemplié ⒻMarcelle Ⓑ93 ⓁS Ⓚch ⓄE dactylo Ⓟ18376"},
                {"text": "Ⓢd ⒻCharles Ⓑ11 ⓁP ⒸC ⓀF Ⓞd Ⓟ14 31"},
                {"text": "ⓈNaudin ⒻMarie Ⓑ53 ⓁS Ⓒv ⓀBelle mère"},
            ],
        ),
    ),
)
def test_run_prediction_language_model(
    image_names,
    language_model_weight,
    expected_predictions,
    tmp_path,
):
    # Make tmpdir and copy needed images inside
    image_dir = tmp_path / "images"
    image_dir.mkdir()
    for image_name in image_names:
        shutil.copyfile(
            (PREDICTION_DATA_PATH / "images" / image_name).with_suffix(".png"),
            (image_dir / image_name).with_suffix(".png"),
        )

    # Update language_model_weight in parameters.yml
    model_path = tmp_path / "models"
    model_path.mkdir(exist_ok=True)

    shutil.copyfile(
        PREDICTION_DATA_PATH / "model.pt",
        model_path / "model.pt",
    )
    shutil.copyfile(
        PREDICTION_DATA_PATH / "charset.pkl",
        model_path / "charset.pkl",
    )

    params = read_yaml(PREDICTION_DATA_PATH / "parameters.yml")
    params["parameters"]["language_model"]["weight"] = language_model_weight
    yaml.dump(params, (model_path / "parameters.yml").open("w"))

    run_prediction(
        image_dir=image_dir,
        font=FONT_PATH,
        maximum_font_size=MAXIMUM_FONT_SIZE,
        model=model_path,
        output=tmp_path,
        confidence_score=False,
        confidence_score_levels=[],
        attention_map=[],
        attention_map_level=None,
        attention_map_scale=0.5,
        alpha_factor=0.9,
        color_map="nipy_spectral",
        word_separators=[" ", "\n"],
        line_separators=["\n"],
        temperature=1.0,
        predict_objects=False,
        max_object_height=None,
        image_extension=".png",
        gpu_device=None,
        batch_size=1,
        tokens=parse_tokens(PREDICTION_DATA_PATH / "tokens.yml"),
        start_token=None,
        use_language_model=True,
        compile_model=False,
        dynamic_mode=False,
    )

    for image_name, expected_prediction in zip(image_names, expected_predictions):
        prediction = json.loads(
            (tmp_path / image_name).with_suffix(".json").read_text()
        )
        assert prediction["text"] == expected_prediction["text"]

        if language_model_weight > 0:
            assert (
                prediction["language_model"]["text"]
                == expected_prediction["language_model"]["text"]
            )
            assert np.isclose(
                prediction["language_model"]["confidence"],
                expected_prediction["language_model"]["confidence"],
            )



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import json
import pickle
import re
from operator import methodcaller

import pytest

from dan.datasets.language_model.build import LanguageModelBuilder
from dan.utils import parse_tokens
from tests import FIXTURES, change_split_content

EXTRACTION_DATA_PATH = FIXTURES / "extraction"

ENTITY_TOKEN_SPACE = re.compile(r"[ⓢ|ⓕ|ⓑ] ")
TWO_SPACES_LM_REGEX = re.compile(r"▁ ▁")


@pytest.mark.parametrize(
    "load_entities,transcription_entities_worker_version,expected_subword_language_corpus",
    (
        (
            True,
            "worker_version_id",
            """▁ ⓢ l a u l ont ▁ ⓕ f r an c oi s ▁ ⓑ 8
▁ ⓢ c i re t ▁ ⓕ an t oi ne ▁ ⓑ 2 7
▁ ⓢ c i re t ▁ ⓕ m a r ie ▁ ⓑ 2 8
▁ ⓢ c i re t ▁ ⓕ m a r ie ▁ ⓑ 2
▁ ⓢ e u re s t on ▁ ⓕ so l an g e ▁ ⓑ 1 0
▁ ⓢ t e r ont u s s ie u x ▁ ⓕ j e an ▁ ⓑ 2
▁ ⓢ p re s s on e t ▁ ⓕ m a r ie ▁ ⓑ 1 2""",
        ),
        (
            False,
            "worker_version_id",
            """▁ la u l ont ▁ f r an c oi s ▁ 8
▁ c i re t ▁ an t oi ne ▁ 2 7
▁ c i re t ▁ m a r ie ▁ 2 8
▁ c i re t ▁ m a r ie ▁ 2
▁ e u res t on ▁ so l an g e ▁ 1 0
▁ t e r ont u ss ie u x ▁ j e an ▁ 2
▁ p res so ne t ▁ m a r ie ▁ 1 2""",
        ),
        (
            True,
            False,
            """▁ ⓢ L a u l o n t ▁ ⓕ F r a n c o i s ▁ ⓑ 8
▁ ⓢ C i r e t ▁ ⓕ A n t o i n e ▁ ⓑ 2 7
▁ ⓢ C i r e t ▁ ⓕ M a r ie ▁ ⓑ 2 8
▁ ⓢ C i r e t ▁ ⓕ M a r ie ▁ ⓑ 2
▁ ⓢ E u r e s t o n ▁ ⓕ S o l a n g e ▁ ⓑ 1 0
▁ ⓢ T e r o n t u s s ie u x ▁ ⓕ J e a n ▁ ⓑ 2
▁ ⓢ P r e s s o n e t ▁ ⓕ M a r ie ▁ ⓑ 1 2""",
        ),
        (
            False,
            False,
            """▁ L a u l ont ▁ F r an c oi s ▁ 8
▁ C i re t ▁ A n t oi n e ▁ 2 7
▁ C i re t ▁ M a r ie ▁ 2 8
▁ C i re t ▁ M a r ie ▁ 2
▁ E u re s t on ▁ S o l an g e ▁ 1 0
▁ T e r ont u s s ie u x ▁ J e an ▁ 2
▁ P re s s on e t ▁ M a r ie ▁ 1 2""",
        ),
    ),
)
@pytest.mark.parametrize("keep_spaces", [True, False])
def test_language_model(
    load_entities,
    keep_spaces,
    transcription_entities_worker_version,
    expected_subword_language_corpus,
    split_content,
    tmp_path,
):
    output = tmp_path / "build"
    (output / "language_model").mkdir(parents=True, exist_ok=True)

    # Mock tokens
    tokens_path = EXTRACTION_DATA_PATH / "tokens.yml"
    tokens = [
        token
        for entity_type in parse_tokens(tokens_path).values()
        for token in [entity_type.start, entity_type.end]
        if token
    ]

    # Mock "labels.json"
    _, labels_content = change_split_content(
        load_entities,
        transcription_entities_worker_version,
        keep_spaces,
        split_content,
        tokens,
        {
            "test": {
                "images/test/dataset_id/test-page_1-line_1.jpg": "ⓢLeunaut  ⓕClau⁇e  ⓑ⁇⁇",
                "images/test/dataset_id/test-page_1-line_2.jpg": "ⓢ⁇aurac⁇o  ⓕClau⁇ine  ⓑ⁇⁇",
                "images/test/dataset_id/test-page_1-line_3.jpg": "ⓢLaurent  ⓕJac⁇use  ⓑ21",
                "images/test/dataset_id/test-page_2-line_1.jpg": "ⓢ⁇alette  ⓕElisa⁇et⁇  ⓑ7⁇",
                "images/test/dataset_id/test-page_2-line_2.jpg": "ⓢTan⁇ol  ⓕJean  ⓑ7⁇",
                "images/test/dataset_id/test-page_2-line_3.jpg": "ⓢ⁇auret  ⓕJean  ⓑ⁇⁇",
            },
            "train": {
                "images/train/dataset_id/train-page_1-line_1.jpg": "ⓢLaulont  ⓕFrancois  ⓑ8",
                "images/train/dataset_id/train-page_1-line_2.jpg": "ⓢCiret  ⓕAntoine  ⓑ27",
                "images/train/dataset_id/train-page_1-line_3.jpg": "ⓢCiret  ⓕMarie  ⓑ28",
                "images/train/dataset_id/train-page_1-line_4.jpg": "ⓢCiret  ⓕMarie  ⓑ2",
                "images/train/dataset_id/train-page_2-line_1.jpg": "ⓢEureston  ⓕSolange  ⓑ10",
                "images/train/dataset_id/train-page_2-line_2.jpg": "ⓢTerontussieux  ⓕJean  ⓑ2",
                "images/train/dataset_id/train-page_2-line_3.jpg": "ⓢPressonet  ⓕMarie  ⓑ12",
            },
            "val": {
                "images/val/dataset_id/val-page_1-line_1.jpg": "ⓢCirau⁇  ⓕAntoine  ⓑ⁇⁇",
                "images/val/dataset_id/val-page_1-line_2.jpg": "ⓢCirau⁇  ⓕPriser  ⓑ⁇⁇",
                "images/val/dataset_id/val-page_1-line_3.jpg": "ⓢCirau⁇  ⓕElisa⁇et⁇  ⓑ⁇⁇",
            },
        },
    )
    (output / "labels.json").write_text(json.dumps(labels_content))

    # Mock "charset.pkl"
    expected_charset = {"⁇"}
    for value in labels_content["train"].values():
        expected_charset.update(set(value))
    if load_entities:
        expected_charset.update(tokens)
    (output / "charset.pkl").write_bytes(pickle.dumps(sorted(list(expected_charset))))

    extractor = LanguageModelBuilder(
        output=output,
        tokens=tokens_path if load_entities else None,
        subword_vocab_size=40,
    )
    extractor.run()

    # Check files
    expected_paths = [
        # Previous files
        output / "charset.pkl",
        output / "labels.json",
        # Language resources
        output / "language_model" / "corpus_characters.txt",
        output / "language_model" / "corpus_subwords.txt",
        output / "language_model" / "corpus_words.txt",
        output / "language_model" / "lexicon_characters.txt",
        output / "language_model" / "lexicon_subwords.txt",
        output / "language_model" / "lexicon_words.txt",
        output / "language_model" / "subword_tokenizer.model",
        output / "language_model" / "subword_tokenizer.vocab",
        output / "language_model" / "tokens.txt",
    ]
    assert sorted(filter(methodcaller("is_file"), output.rglob("*"))) == expected_paths

    # Check "language_corpus.txt"
    expected_char_language_corpus = """ⓢ L a u l o n t ▁ ▁ ⓕ F r a n c o i s ▁ ▁ ⓑ 8
ⓢ C i r e t ▁ ▁ ⓕ A n t o i n e ▁ ▁ ⓑ 2 7
ⓢ C i r e t ▁ ▁ ⓕ M a r i e ▁ ▁ ⓑ 2 8
ⓢ C i r e t ▁ ▁ ⓕ M a r i e ▁ ▁ ⓑ 2
ⓢ E u r e s t o n ▁ ▁ ⓕ S o l a n g e ▁ ▁ ⓑ 1 0
ⓢ T e r o n t u s s i e u x ▁ ▁ ⓕ J e a n ▁ ▁ ⓑ 2
ⓢ P r e s s o n e t ▁ ▁ ⓕ M a r i e ▁ ▁ ⓑ 1 2"""

    expected_word_language_corpus = """ⓢ Laulont ▁ ⓕ Francois ▁ ⓑ 8
ⓢ Ciret ▁ ⓕ Antoine ▁ ⓑ 27
ⓢ Ciret ▁ ⓕ Marie ▁ ⓑ 28
ⓢ Ciret ▁ ⓕ Marie ▁ ⓑ 2
ⓢ Eureston ▁ ⓕ Solange ▁ ⓑ 10
ⓢ Terontussieux ▁ ⓕ Jean ▁ ⓑ 2
ⓢ Pressonet ▁ ⓕ Marie ▁ ⓑ 12"""

    # Transcriptions with worker version are in lowercase
    if transcription_entities_worker_version:
        expected_char_language_corpus = expected_char_language_corpus.lower()
        expected_word_language_corpus = expected_word_language_corpus.lower()
        expected_subword_language_corpus = expected_subword_language_corpus.lower()

    # If we do not load entities, remove tokens
    if not load_entities:
        expected_char_language_corpus = ENTITY_TOKEN_SPACE.sub(
            "", expected_char_language_corpus
        )
        expected_word_language_corpus = ENTITY_TOKEN_SPACE.sub(
            "", expected_word_language_corpus
        )
        expected_subword_language_corpus = ENTITY_TOKEN_SPACE.sub(
            "", expected_subword_language_corpus
        )
    # Replace double spaces with regular space
    if not keep_spaces:
        expected_char_language_corpus = TWO_SPACES_LM_REGEX.sub(
            "▁", expected_char_language_corpus
        )
        expected_word_language_corpus = TWO_SPACES_LM_REGEX.sub(
            "▁", expected_word_language_corpus
        )
        expected_subword_language_corpus = TWO_SPACES_LM_REGEX.sub(
            "▁", expected_subword_language_corpus
        )

    assert (
        output / "language_model" / "corpus_characters.txt"
    ).read_text() == expected_char_language_corpus

    assert (
        output / "language_model" / "corpus_words.txt"
    ).read_text() == expected_word_language_corpus

    assert (
        output / "language_model" / "corpus_subwords.txt"
    ).read_text() == expected_subword_language_corpus

    # Check "language_tokens.txt"
    expected_language_tokens = [
        "▁" if t.isspace() else t for t in sorted(list(expected_charset))
    ]
    expected_language_tokens.append("◌")
    assert (output / "language_model" / "tokens.txt").read_text() == "\n".join(
        expected_language_tokens
    )

    # Check "language_lexicon.txt"
    expected_language_char_lexicon = [f"{t} {t}" for t in expected_language_tokens]
    assert (
        output / "language_model" / "lexicon_characters.txt"
    ).read_text() == "\n".join(expected_language_char_lexicon)

    word_vocab = set([word for word in expected_word_language_corpus.split()])
    expected_language_word_lexicon = [
        f"{word} {' '.join(word)}" for word in sorted(word_vocab)
    ]
    assert (output / "language_model" / "lexicon_words.txt").read_text() == "\n".join(
        expected_language_word_lexicon
    )

    subword_vocab = set(
        [subword for subword in expected_subword_language_corpus.split()]
    )
    expected_language_subword_lexicon = [
        f"{subword} {' '.join(subword)}" for subword in sorted(subword_vocab)
    ]
    assert (
        output / "language_model" / "lexicon_subwords.txt"
    ).read_text() == "\n".join(expected_language_subword_lexicon)


@pytest.mark.parametrize(
    "expected_subword_language_corpus,subword_vocab_size",
    (
        (
            """▁ ⓢ L a u l o n t ▁ ⓕ F r a n c o i s ▁ ⓑ 8
▁ ⓢ C i r e t ▁ ⓕ A n t o i n e ▁ ⓑ 2 7
▁ ⓢ C i r e t ▁ ⓕ M a r ie ▁ ⓑ 2 8
▁ ⓢ C i r e t ▁ ⓕ M a r ie ▁ ⓑ 2
▁ ⓢ E u r e s t o n ▁ ⓕ S o l a n g e ▁ ⓑ 1 0
▁ ⓢ T e r o n t u s s ie u x ▁ ⓕ J e a n ▁ ⓑ 2
▁ ⓢ P r e s s o n e t ▁ ⓕ M a r ie ▁ ⓑ 1 2""",
            40,
        ),
        (
            """▁ ⓢ L a u l ont ▁ ⓕ F r an c oi s ▁ ⓑ 8
▁ ⓢ C i re t ▁ ⓕ A n t oi n e ▁ ⓑ 2 7
▁ ⓢ C i re t ▁ ⓕ M a r ie ▁ ⓑ 2 8
▁ ⓢ C i re t ▁ ⓕ M a r ie ▁ ⓑ 2
▁ ⓢ E u re s t on ▁ ⓕ S o l an g e ▁ ⓑ 1 0
▁ ⓢ T e r ont u s s ie u x ▁ ⓕ J e an ▁ ⓑ 2
▁ ⓢ P re s s on e t ▁ ⓕ M a r ie ▁ ⓑ 1 2""",
            45,
        ),
    ),
)
@pytest.mark.parametrize("keep_spaces", [True, False])
def test_language_model_subword_vocab_size(
    keep_spaces,
    expected_subword_language_corpus,
    subword_vocab_size,
    split_content,
    tmp_path,
):
    output = tmp_path / "build"
    (output / "language_model").mkdir(parents=True, exist_ok=True)

    # Mock tokens
    tokens_path = EXTRACTION_DATA_PATH / "tokens.yml"
    tokens = [
        token
        for entity_type in parse_tokens(tokens_path).values()
        for token in [entity_type.start, entity_type.end]
        if token
    ]

    # Mock "labels.json"
    _, labels_content = change_split_content(
        True,
        False,
        keep_spaces,
        split_content,
        tokens,
        {
            "test": {
                "images/test/dataset_id/test-page_1-line_1.jpg": "ⓢLeunaut  ⓕClau⁇e  ⓑ⁇⁇",
                "images/test/dataset_id/test-page_1-line_2.jpg": "ⓢ⁇aurac⁇o  ⓕClau⁇ine  ⓑ⁇⁇",
                "images/test/dataset_id/test-page_1-line_3.jpg": "ⓢLaurent  ⓕJac⁇use  ⓑ21",
                "images/test/dataset_id/test-page_2-line_1.jpg": "ⓢ⁇alette  ⓕElisa⁇et⁇  ⓑ7⁇",
                "images/test/dataset_id/test-page_2-line_2.jpg": "ⓢTan⁇ol  ⓕJean  ⓑ7⁇",
                "images/test/dataset_id/test-page_2-line_3.jpg": "ⓢ⁇auret  ⓕJean  ⓑ⁇⁇",
            },
            "train": {
                "images/train/dataset_id/train-page_1-line_1.jpg": "ⓢLaulont  ⓕFrancois  ⓑ8",
                "images/train/dataset_id/train-page_1-line_2.jpg": "ⓢCiret  ⓕAntoine  ⓑ27",
                "images/train/dataset_id/train-page_1-line_3.jpg": "ⓢCiret  ⓕMarie  ⓑ28",
                "images/train/dataset_id/train-page_1-line_4.jpg": "ⓢCiret  ⓕMarie  ⓑ2",
                "images/train/dataset_id/train-page_2-line_1.jpg": "ⓢEureston  ⓕSolange  ⓑ10",
                "images/train/dataset_id/train-page_2-line_2.jpg": "ⓢTerontussieux  ⓕJean  ⓑ2",
                "images/train/dataset_id/train-page_2-line_3.jpg": "ⓢPressonet  ⓕMarie  ⓑ12",
            },
            "val": {
                "images/val/dataset_id/val-page_1-line_1.jpg": "ⓢCirau⁇  ⓕAntoine  ⓑ⁇⁇",
                "images/val/dataset_id/val-page_1-line_2.jpg": "ⓢCirau⁇  ⓕPriser  ⓑ⁇⁇",
                "images/val/dataset_id/val-page_1-line_3.jpg": "ⓢCirau⁇  ⓕElisa⁇et⁇  ⓑ⁇⁇",
            },
        },
    )
    (output / "labels.json").write_text(json.dumps(labels_content))

    # Mock "charset.pkl"
    expected_charset = {"⁇"}
    for value in labels_content["train"].values():
        expected_charset.update(set(value))
    expected_charset.update(tokens)
    (output / "charset.pkl").write_bytes(pickle.dumps(sorted(list(expected_charset))))

    extractor = LanguageModelBuilder(
        output=output,
        tokens=tokens_path,
        subword_vocab_size=subword_vocab_size,
    )
    extractor.run()

    # Check files
    expected_paths = [
        # Previous files
        output / "charset.pkl",
        output / "labels.json",
        # Language resources
        output / "language_model" / "corpus_characters.txt",
        output / "language_model" / "corpus_subwords.txt",
        output / "language_model" / "corpus_words.txt",
        output / "language_model" / "lexicon_characters.txt",
        output / "language_model" / "lexicon_subwords.txt",
        output / "language_model" / "lexicon_words.txt",
        output / "language_model" / "subword_tokenizer.model",
        output / "language_model" / "subword_tokenizer.vocab",
        output / "language_model" / "tokens.txt",
    ]
    assert sorted(filter(methodcaller("is_file"), output.rglob("*"))) == expected_paths

    # Check "language_corpus.txt"
    expected_char_language_corpus = """ⓢ L a u l o n t ▁ ▁ ⓕ F r a n c o i s ▁ ▁ ⓑ 8
ⓢ C i r e t ▁ ▁ ⓕ A n t o i n e ▁ ▁ ⓑ 2 7
ⓢ C i r e t ▁ ▁ ⓕ M a r i e ▁ ▁ ⓑ 2 8
ⓢ C i r e t ▁ ▁ ⓕ M a r i e ▁ ▁ ⓑ 2
ⓢ E u r e s t o n ▁ ▁ ⓕ S o l a n g e ▁ ▁ ⓑ 1 0
ⓢ T e r o n t u s s i e u x ▁ ▁ ⓕ J e a n ▁ ▁ ⓑ 2
ⓢ P r e s s o n e t ▁ ▁ ⓕ M a r i e ▁ ▁ ⓑ 1 2"""

    expected_word_language_corpus = """ⓢ Laulont ▁ ⓕ Francois ▁ ⓑ 8
ⓢ Ciret ▁ ⓕ Antoine ▁ ⓑ 27
ⓢ Ciret ▁ ⓕ Marie ▁ ⓑ 28
ⓢ Ciret ▁ ⓕ Marie ▁ ⓑ 2
ⓢ Eureston ▁ ⓕ Solange ▁ ⓑ 10
ⓢ Terontussieux ▁ ⓕ Jean ▁ ⓑ 2
ⓢ Pressonet ▁ ⓕ Marie ▁ ⓑ 12"""

    # Replace double spaces with regular space
    if not keep_spaces:
        expected_char_language_corpus = TWO_SPACES_LM_REGEX.sub(
            "▁", expected_char_language_corpus
        )
        expected_word_language_corpus = TWO_SPACES_LM_REGEX.sub(
            "▁", expected_word_language_corpus
        )
        expected_subword_language_corpus = TWO_SPACES_LM_REGEX.sub(
            "▁", expected_subword_language_corpus
        )

    assert (
        output / "language_model" / "corpus_characters.txt"
    ).read_text() == expected_char_language_corpus

    assert (
        output / "language_model" / "corpus_words.txt"
    ).read_text() == expected_word_language_corpus

    assert (
        output / "language_model" / "corpus_subwords.txt"
    ).read_text() == expected_subword_language_corpus

    # Check "language_tokens.txt"
    expected_language_tokens = [
        "▁" if t.isspace() else t for t in sorted(list(expected_charset))
    ]
    expected_language_tokens.append("◌")
    assert (output / "language_model" / "tokens.txt").read_text() == "\n".join(
        expected_language_tokens
    )

    # Check "language_lexicon.txt"
    expected_language_char_lexicon = [f"{t} {t}" for t in expected_language_tokens]
    assert (
        output / "language_model" / "lexicon_characters.txt"
    ).read_text() == "\n".join(expected_language_char_lexicon)

    word_vocab = set([word for word in expected_word_language_corpus.split()])
    expected_language_word_lexicon = [
        f"{word} {' '.join(word)}" for word in sorted(word_vocab)
    ]
    assert (output / "language_model" / "lexicon_words.txt").read_text() == "\n".join(
        expected_language_word_lexicon
    )

    subword_vocab = set(
        [subword for subword in expected_subword_language_corpus.split()]
    )
    expected_language_subword_lexicon = [
        f"{subword} {' '.join(subword)}" for subword in sorted(subword_vocab)
    ]
    assert (
        output / "language_model" / "lexicon_subwords.txt"
    ).read_text() == "\n".join(expected_language_subword_lexicon)



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import itertools
import logging
import operator
from dataclasses import dataclass, field
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import List

import sentencepiece as spm
from nltk import wordpunct_tokenize

from dan.utils import EntityType, LMTokenMapping

logger = logging.getLogger(__name__)


def get_vocabulary(tokenized_text: List[str]) -> set[str]:
    """
    Compute set of vocabulary from tokenzied text.
    :param tokenized_text: List of tokenized text.
    """
    return sorted(set([token for doc in tokenized_text for token in doc.split()]))


@dataclass
class Tokenizer:
    """
    A multi-level tokenizer (char, subword, word), where the subword tokenizer is trained using sentencepiece.
    :param training_corpus: List of training text.
    :param outdir: Path to save the subword tokenizer.
    :param mapping: Mapping between displayed and encoded versions of special characters.
    :param tokens: Start and end tokens used to represent named entities.
    :param subword_vocab_size: Size of the vocabulary size to use to train the subword tokenizer.
    """

    training_corpus: List[str]
    charset: List[str]
    unknown_token: str
    outdir: Path
    mapping: LMTokenMapping
    tokens: EntityType | None = None
    subword_vocab_size: int = 1000
    sentencepiece_model: spm.SentencePieceProcessor = field(init=False)

    @property
    def prefix(self) -> Path:
        return self.outdir / "subword_tokenizer"

    @property
    def ner_tokens(self) -> List[str]:
        if self.tokens is None:
            return []
        return list(
            itertools.chain(
                map(operator.attrgetter("start"), self.tokens.values()),
                filter(
                    operator.truth,
                    map(operator.attrgetter("end"), self.tokens.values()),
                ),
            )
        )

    @property
    def mapping_tokens(self) -> List[str]:
        return [token.encoded for token in self.mapping]

    @property
    def special_tokens(self) -> List[str]:
        return list(set(itertools.chain(self.mapping_tokens, self.ner_tokens)))

    def __post_init__(self) -> None:
        """
        Train a sentencepiece model on the training corpus.
        """
        # Write the corpus in a text file
        logger.info("Training a sentencepiece model for subword tokenization")
        with NamedTemporaryFile(dir=self.outdir, suffix=".txt", mode="w") as tmp_file:
            tmp_file.write("\n".join(self.training_corpus))
            tmp_file.flush()

            try:
                spm.SentencePieceTrainer.train(
                    input=tmp_file.name,
                    vocab_size=self.subword_vocab_size,
                    model_prefix=self.prefix,
                    user_defined_symbols=self.special_tokens,
                    minloglevel=1,
                )
            except Exception as e:
                logger.warning(
                    f"Failed to train a sentencepiece model for subword tokenization: {e} "
                    "Try again by editing the `--subword-vocab-size` parameter."
                )
                self.sentencepiece_model = None
                return

        # Load the model
        self.sentencepiece_model = spm.SentencePieceProcessor(
            model_file=str(self.prefix.with_suffix(".model"))
        )

    def subword_tokenize(self, text: str) -> str:
        """
        Tokenize into subwords. Sampling is disabled to ensure reproducibility.
        """
        tokens = self.sentencepiece_model.encode(text, out_type=str)
        return " ".join(map("".join, map(self.encode, tokens)))

    def word_tokenize(self, text: str) -> str:
        """
        Tokenize text into a string of space-separated words. Spaces (⎵) and NER tokens are considered as words.
        :param text: Text to be tokenized.
        """
        words = list(map("".join, map(self.encode, wordpunct_tokenize(text))))
        return " ".join(
            [
                f"{word} {self.mapping.space.encoded}"
                if (i != len(words) - 1 and word not in self.ner_tokens)
                else word
                for i, word in enumerate(words)
            ]
        )

    def char_tokenize(self, text: str) -> str:
        """
        Tokenize text into a string of space-separated characters.
        :param text: Text to be tokenized.
        """
        return " ".join(
            [
                char if char in self.charset else self.unknown_token
                for char in self.encode(text)
            ]
        )

    def encode(self, text: List[str]) -> List[str]:
        """
        Encode special tokens.
        :param text: Text to be encoded.
        """
        return map(self.mapping.encode_token, text)



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
from pathlib import Path


class ImageDownloadError(Exception):
    """
    Raised when an element's image could not be downloaded
    """

    def __init__(
        self, split: str, path: Path, url: str, exc: Exception, *args: object
    ) -> None:
        super().__init__(*args)
        self.split: str = split
        self.path: str = str(path)
        self.url: str = url
        self.message = f"{str(exc)} for element {path.stem}"



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
"""
Preprocess datasets for training.
"""

from dan.datasets.analyze import add_analyze_parser
from dan.datasets.download import add_download_parser
from dan.datasets.entities import add_entities_parser
from dan.datasets.extract import add_extract_parser
from dan.datasets.language_model import add_language_model_parser
from dan.datasets.tokens import add_tokens_parser


def add_dataset_parser(subcommands) -> None:
    parser = subcommands.add_parser(
        "dataset",
        description=__doc__,
        help=__doc__,
    )
    subcommands = parser.add_subparsers(metavar="subcommand")

    add_extract_parser(subcommands)
    add_download_parser(subcommands)
    add_analyze_parser(subcommands)
    add_entities_parser(subcommands)
    add_tokens_parser(subcommands)
    add_language_model_parser(subcommands)



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
from operator import itemgetter
from pathlib import Path

import yaml

from arkindex_export import EntityType, open_database


def run(database: Path, output_file: Path) -> None:
    # Load SQLite database
    open_database(database)

    # Extract and save entities to YAML
    entities = list(
        map(itemgetter(0), EntityType.select(EntityType.name).distinct().tuples())
    )

    output_file.parent.mkdir(exist_ok=True, parents=True)
    output_file.write_text(
        yaml.safe_dump({"entities": entities}, explicit_start=True, allow_unicode=True)
    )



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

import json
import logging
import pickle
from collections import defaultdict
from pathlib import Path
from typing import Dict

from dan.datasets.extract.arkindex import TRAIN_NAME
from dan.datasets.language_model.utils import (
    Tokenizer,
    get_vocabulary,
)
from dan.utils import LMTokenMapping, parse_tokens

LANGUAGE_DIR = "language_model"  # Subpath to the language model directory.

logger = logging.getLogger(__name__)


class LanguageModelBuilder:
    """
    Build a language model from extracted data
    """

    def __init__(
        self,
        output: Path | None = None,
        subword_vocab_size: int = 1000,
        unknown_token: str = "⁇",
        tokens: Path | None = None,
    ) -> None:
        self.output = output

        self.unknown_token = unknown_token
        self.tokens = parse_tokens(tokens) if tokens else {}

        self.subword_vocab_size = subword_vocab_size
        self.mapping = LMTokenMapping()

        self.language_corpus = defaultdict(list)
        self.language_tokens = []
        self.language_lexicon = defaultdict(list)

        # Load labels file
        labels_file = self.output / "labels.json" if self.output else None
        self.labels: Dict = (
            json.loads(labels_file.read_text())
            if labels_file and labels_file.is_file()
            else {}
        )

        # Load charset file
        charset_file = self.output / "charset.pkl" if self.output else None
        self.charset: Dict = (
            pickle.loads(charset_file.read_bytes())
            if charset_file and charset_file.is_file()
            else {}
        )

    def format_lm_files(self) -> None:
        """
        Convert charset to a LM-compatible charset. Ensure that special LM tokens do not appear in the charset.
        """
        logger.info("Preparing language resources")

        # Build LM tokens
        for token in sorted(list(self.charset)):
            assert (
                token not in self.mapping.encode.values()
            ), f"Special token {token} is reserved for language modeling."
            self.language_tokens.append(
                self.mapping.encode[token]
            ) if token in self.mapping.encode else self.language_tokens.append(token)
        self.language_tokens.append(self.mapping.ctc.encoded)

        # Build LM corpus
        train_corpus = [
            value.replace(self.mapping.linebreak.display, self.mapping.space.display)
            for value in self.labels[TRAIN_NAME].values()
        ]

        tokenizer = Tokenizer(
            training_corpus=train_corpus,
            charset=self.language_tokens,
            unknown_token=self.unknown_token,
            outdir=self.output / LANGUAGE_DIR,
            mapping=self.mapping,
            tokens=self.tokens,
            subword_vocab_size=self.subword_vocab_size,
        )

        if not tokenizer.sentencepiece_model:
            return

        for level, tokenize in (
            ("characters", tokenizer.char_tokenize),
            ("words", tokenizer.word_tokenize),
            ("subwords", tokenizer.subword_tokenize),
        ):
            self.language_corpus[level] = list(map(tokenize, train_corpus))

        # Build LM lexicon
        self.language_lexicon["characters"] = [
            f"{token} {token}" for token in self.language_tokens
        ]
        for level in ["words", "subwords"]:
            self.language_lexicon[level] = [
                f"{token} {tokenizer.char_tokenize(token)}"
                for token in get_vocabulary(self.language_corpus[level])
            ]

    def export(self) -> None:
        """
        Writes all files needed for the language model
        """
        for level in ["characters", "words", "subwords"]:
            (self.output / LANGUAGE_DIR / f"corpus_{level}.txt").write_text(
                "\n".join(self.language_corpus[level])
            )
            (self.output / LANGUAGE_DIR / f"lexicon_{level}.txt").write_text(
                "\n".join(self.language_lexicon[level])
            )

        (self.output / LANGUAGE_DIR / "tokens.txt").write_text(
            "\n".join(self.language_tokens)
        )

    def run(self) -> None:
        """
        Build and write all files needed for the language model
        """
        self.format_lm_files()
        self.export()


def run(
    output: Path,
    subword_vocab_size: int,
    unknown_token: str,
    tokens: Path | None,
):
    """
    Build and write all files needed for the language model

    :param output: Path where the `split.json` file is stored and where the data will be generated
    :param subword_vocab_size: The size of the subword vocabulary.
    :param unknown_token: The token used to replace unknown characters.
    :param tokens: Mapping between starting tokens and end tokens to extract text with their entities.
    """
    (output / LANGUAGE_DIR).mkdir(parents=True, exist_ok=True)

    LanguageModelBuilder(
        output=output,
        subword_vocab_size=subword_vocab_size,
        unknown_token=unknown_token,
        tokens=tokens,
    ).run()



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
"""
Download images of a dataset from a split extracted by DAN
"""

import pathlib

from dan.datasets.download.images import run


def _valid_image_format(value: str):
    im_format = value
    if not im_format.startswith("."):
        im_format = "." + im_format
    return im_format


def add_download_parser(subcommands) -> None:
    parser = subcommands.add_parser(
        "download",
        description=__doc__,
        help=__doc__,
    )

    # Required arguments.
    parser.add_argument(
        "--output",
        type=pathlib.Path,
        help="Path where the `split.json` file is stored and where the data will be generated.",
        required=True,
    )

    parser.add_argument(
        "--max-width",
        type=int,
        help="Images larger than this width will be resized to this width.",
    )

    parser.add_argument(
        "--max-height",
        type=int,
        help="Images larger than this height will be resized to this height.",
    )

    # Formatting arguments
    parser.add_argument(
        "--image-format",
        type=_valid_image_format,
        default=".jpg",
        help="Images will be saved under this format.",
    )

    parser.add_argument(
        "--unknown-token",
        type=str,
        default="⁇",
        help="Token to use to replace character in the validation/test sets that is not included in the training set.",
    )

    parser.set_defaults(func=run)
